# socklen_t.m4
#
# Copyright (C) 2005-2008 Free Software Foundation, Inc.
# This file is free software; the Free Software Foundation
# gives unlimited permission to copy and/or distribute it,
# with or without modifications, as long as this notice is preserved.
#
# From Albert Chin, Windows fixes from Simon Josefsson.
# autoconf 2.62 rewrite + fixes by Ger Hobbelt, 2008
#
# Check for socklen_t: historically on BSD it is an int, and in
# POSIX 1g it is a type of its own, but some platforms use different
# types for the argument to getsockopt, getpeername, etc.  So we
# have to test to find something that will work.
#
# On mingw32, socklen_t is in ws2tcpip.h ('int'), so we try to find
# it there first.  That file is included by gnulib's socket_.h, which
# all users of this module should include.  Cygwin must not include
# ws2tcpip.h.
#



# (this macro for local use only:)
# AX_SOCKLEN_PREREQUISITES exists to make sure the checks in here do not
# mix their console output with AX_TYPE_SOCKLEN_T when one or more of these
# items needs to be performed.
AC_DEFUN([AX_SOCKLEN_PREREQUISITES],
[
  AC_CHECK_HEADERS([sys/types.h sys/socket.h ws2tcpip.h])
#  AC_CHECK_FUNCS([getpeername])
])


# AX_TYPE_SOCKLEN_T
# ---------------------
AC_DEFUN([AX_TYPE_SOCKLEN_T],
[
  AC_REQUIRE([AX_SOCKLEN_PREREQUISITES])
  AC_CACHE_CHECK([for socklen_t], [ac_cv_type_socklen_t],
  [
    AC_COMPILE_IFELSE([
      AC_LANG_PROGRAM(
[AC_INCLUDES_DEFAULT([])[

@%:@if HAVE_SYS_TYPES_H
@%:@include <sys/types.h>
@%:@endif
@%:@if HAVE_SYS_SOCKET_H
@%:@ include <sys/socket.h>
@%:@elif HAVE_WS2TCPIP_H
@%:@ include <ws2tcpip.h>
@%:@endif

]],[[

    socklen_t t = 1;
    return 0;
]]
    )],
    [ac_cv_type_socklen_t=yes],
    [ac_cv_type_socklen_t=no])
  ])
  if test "x${ac_cv_type_socklen_t}" = "xyes"; then
    AC_DEFINE([HAVE_SOCKLEN_T], 1,
      [Define to 1 if the system has the type `socklen_t'.])
  else
    AC_MSG_CHECKING([for socklen_t equivalent])
    AC_CACHE_VAL([ax_cv_ax_cv_socklen_t_equiv],
    [
      ax_cv_socklen_t_equiv=''
      for t in int size_t "unsigned int" "long int" "unsigned long int"
      do
        # compiling code using getpeername(0,0,&typed_var) is bloody useless, as
        # at least gcc only gives a WARNING instead of an error!
        # So we do some run-time sizeof and signed-ness checking instead...
        AC_RUN_IFELSE([
          AC_LANG_PROGRAM(
[AC_INCLUDES_DEFAULT([])[

@%:@if HAVE_SYS_TYPES_H
@%:@include <sys/types.h>
@%:@endif
@%:@if HAVE_SYS_SOCKET_H
@%:@ include <sys/socket.h>
@%:@elif HAVE_WS2TCPIP_H
@%:@ include <ws2tcpip.h>
@%:@endif

]],[[
    $t len;
    struct msghdr cmh;
    int debug = 0;

    cmh.msg_namelen = -1L; /* ignore signed/unsigned/size warnings. */
    len = -1L; /* ignore signed/unsigned/size warnings. */

    if (debug)
    {
        printf(" !(sizeof(cmh.msg_namelen) == sizeof(len)) : %d vs %d = %ld vs %ld\n",
                (int)sizeof(cmh.msg_namelen), (int)sizeof(len),
                (long int)cmh.msg_namelen, (long int)len);
    }
    if (sizeof(cmh.msg_namelen) == sizeof(len))
    {
      /*
         check signedness: the trick is that the comparison will be
         'promoted' to the widest type (including signed --> unsigned)
         according to C89/99 rules and THAT means an UNSIGNED '-1' bit
         pattern will NOT be considered less than the arbitrary >0 value
         '16' here:
       */
      if (cmh.msg_namelen < 16 && len < 16)
      {
        /* both signed type! */
        if (debug) printf("signed!\n");
        return 0;
      }
      else if (cmh.msg_namelen > 16 && len > 16)
      {
        /* both UNsigned type! */
        if (debug) printf("UNsigned!\n");
        return 0;
      }
      /* different sign for both types */
      return 1;
    }
    /* different size: */
    return 2;
]]
        )],
        [
          ax_cv_socklen_t_equiv="$t"
          break
        ],
        [])
      done
    ])
    if test "$ax_cv_socklen_t_equiv" = ""; then
      AC_MSG_ERROR([Cannot find a type to use in place of socklen_t])
    else
      AC_MSG_RESULT([$ax_cv_socklen_t_equiv])
      AC_DEFINE_UNQUOTED([socklen_t], [$ax_cv_socklen_t_equiv],
       [type to use in place of socklen_t if not defined])
    fi
  fi
])

