/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "myssl.h"
#include "gui_api.h"
#include "net.h"                /* [i_a] */
#include "absio.h"
#include "tools.h"



/*
 * Return 1 when the 'sock' connection has data available for
 * reading.
 *
 * Return 0 when the timeout (> 0) [unit: msec] has expired.
 *
 * Return -1 when an error, stop or break has occurred.
 *
 *
 * CAVEAT: tl_selectr() will return -1 when an 'exception' occurs according
 *         to select(,, &exception_fdset) -- see UNIX/... man page when
 *         this may happen (OOB data, for instance).
 *
 *
 * Not that this is to be considered as 'standard', but here's what Microsoft
 * has to say about select() -- please note the intricacies about 'exceptional
 * error conditions' and such -- :
 *
 *   The parameter exceptfds identifies the sockets that are to be checked for
 *   the presence of OOB data or any exceptional error conditions.
 *
 *   Note  Out-of-band data will only be reported in this way if the option
 *         SO_OOBINLINE is FALSE. If a socket is processing a connect
 *         call (nonblocking), failure of the connect attempt is indicated in
 *         exceptfds (application must then call getsockopt SO_ERROR to determine
 *         the error value to describe why the failure occurred). This document
 *         does not define which other errors will be included.
 *
 *   [...]
 *
 *   In summary, a socket will be identified in a particular set when select
 *   returns if:
 *
 *   readfds:
 *
 *     If listen has been called and a connection is pending, accept will succeed.
 *     Data is available for reading (includes OOB data if SO_OOBINLINE is enabled).
 *     Connection has been closed/reset/terminated.
 *
 *   writefds:
 *
 *     If processing a connect call (nonblocking), connection has succeeded.
 *     Data can be sent.
 *
 *   exceptfds:
 *
 *     If processing a connect call (nonblocking), connection attempt failed.
 *     OOB data is available for reading (only if SO_OOBINLINE is disabled).
 *
 * End of MS quote.
 *
 *
 * CAVEAT EMPTOR ONCE MORE!
 *
 * Given the above, how about an incoming CLOSE request while we DO NOT watch the
 * READ set as we are only interested in WRITE (tl_selectw() that is!)?
 *
 * Well, luck has that we have two possible scenarios there:
 *
 * 1) The server sent us a FULL CLOSE request as things went pear shaped over
 * there. No sweat, because a subsequent WRITE() attempt will catch that
 * one. BUT, but, but... when our dear pavuk user has not specified a NONZERO
 * timeout, we WILL NEVER LEAVE the select() loop in tl_selectw()... or will
 * we? I haven't been able to thoroughly test all possible flavors of this, but
 * IT LOOKS LIKE 'the usual suspects' ALSO signal the WRITE FD under this
 * circumstance. It's however a wicked case as it will only screw you over when
 * the transmit queue is filled to the gills while a CLOSE request comes in from
 * the server(!) (we are playing client/connect, except in the active FTP case for
 * our data connection) AND we just entered select(,WRITE,EXCEPT) (== tl_selectw()).
 *
 * So should we play ultra-safe and check for a READ FD signal too in tl_selectw()?
 * That would complicate matters - okay, HTTP and the other stuff supported by pavuk
 * is essentially sorta half-duplex, so this won't bite us, unless we're in a
 * 'persistent' connection. HTTP/1.1 servers can do that to us...
 *
 *
 * what if you just generally you've been able to
 * push all data out the door, despite the fact
 * as we: the select() you will find that we need to include READ checking for a
 * select() on WRITE too (tl_selectw()) as we would otherwise MISS those server-side
 * connection close requests when things go pear-shaped and our dear pavuk user
 * did not specify a non-zero 'write timeout'.
 *
 * The trouble in THAT particular case (we CANNOT see if it's either
 * 'data pending for READ' OR a 'requested CLOSE' (how about a single-sided
 * shutdown() to liven things up, hmmm?) with tl_selectw() is that CANNOT simply
 * say a 'READ' signal there is an error as we can do for FD_ISSET() on the EXCEPT
 * set. So we'd better return a SPECIAL VALUE there and pray the code using
 * tl_selectw() can cope with that.
 *
 * The 'special value'?
 *
 *   -2
 *
 * But only for tl_selectw() and a PROBABLE CLOSE request pending.
 *
 *
 *
 * WARNING: comment must be updated as it is VERY WRONG about many things!
 * WARNING: comment must be updated as it is VERY WRONG about many things!
 * WARNING: comment must be updated as it is VERY WRONG about many things!
 * WARNING: comment must be updated as it is VERY WRONG about many things!
 * WARNING: comment must be updated as it is VERY WRONG about many things!
 *
 *
 *
 */
int tl_selectr(int sock, long int timeout)
{
    fd_set readfds, exceptfds;
    struct timeval tout, *toutptr = NULL;
    int rv;

    /* [i_a] */
    FD_ZERO(&readfds);
    FD_ZERO(&exceptfds);

    /*
     * FD_SET will crash if you pass it sock = -1
     */
    ASSERT(sock_maybe_in_fdset(&readfds, sock));
    if (!sock_maybe_in_fdset(&readfds, sock))
    {
        char *msg = tl_str_printf(NULL, "(sock = %d)", sock);

        errno = EINVAL;
        xperror1("tl_selectr:FD_SETSIZE ", msg);
        _free(msg);
        return -1;
    }

    FD_SET(sock, &readfds);
    FD_SET(sock, &exceptfds);

#ifdef HAVE_WINSOCK2_H
    WSASetLastError(0);
#endif
    errno = 0;
    ASSERT(timeout >= 0);
    if (timeout > 0)
    {
        tout.tv_sec = timeout / 1000;
        tout.tv_usec = timeout % 1000;
        toutptr = &tout;
        rv = select(sock + 1, &readfds, NULL, &exceptfds, toutptr);
    }
    else
    {
        /* we could've used pselect() here, but not all systems support it [properly] */
        do
        {
            /* see comment below: must be re-initialized during EVERY round! */
            FD_ZERO(&readfds);
            FD_ZERO(&exceptfds);
            FD_SET(sock, &readfds);
            FD_SET(sock, &exceptfds);

            tout.tv_sec = 1;
            tout.tv_usec = 0;
            toutptr = &tout;
            rv = select(sock + 1, &readfds, NULL, &exceptfds, toutptr);
            DEBUG_DEVEL(DBGvars("select(sock = %d) --> read: %d / exception: %d / return: %d\n"), sock, (int)FD_ISSET(sock, &readfds),
                        (int)FD_ISSET(sock, &exceptfds), rv);
            if (cfg.debug && (DEBUG_NET_BITMASK & cfg.debug_level))
            {
                if ((rv == 0) && !cfg.stop && !cfg.rbreak && !FD_ISSET(sock, &readfds))
                {
                    char buf[32];
                    int r;
                    char buf2[64];

                    r = recv(sock, buf, 64, MSG_PEEK);

                    DEBUG_NET("recv(sock = %d, MSG_PEEK) --> read: %d / errno: %d / content: '%s'\n",
                              sock, r, errno, tl_asciidump_content(buf2, sizeof(buf2), (r > 0 ? buf : "---"), (r > 0 ? r : 3)));

                    /*
                     * [i_a] :-S - Okay, so this happens a few times on Suse10.2/AMD64 and it looks like this
                     * might affect other boxes as well, but come /ON/: recv() shows the TCP/IP stack has bytes
                     * available for reading, while select() does NOT signal this???
                     *
                     * Aiiiiiiiiiiiiiiiiiii!
                     *
                     * *ME* *MORON*!
                     *
                     * Of course! select() RESETs the FD_SET handle when no data is available. Now imagine it takes a
                     * while for the data to arrive (happens to me now that I am testing a severely loaded HTTP server!)
                     * which would lead to the first round (timeout = 1 second!) to RESET that FD_SET(readfds) and
                     * since i did NOT re-setup that FD_SET there, the NEXT round of select(timeout=1s) will NOT see
                     * ANY handles to check... which is perfectly okay from the kernel's viewpoint, but NOT from this
                     * software! We /should/ have re-initialized those FD_SET()s for EVERY round of select()!
                     */
                }
            }
        } while ((rv == 0) && !cfg.stop && !cfg.rbreak);
        /* unlimited time: treat 'timeout' as an error as it can only happen due to stop/break. */
        if (rv == 0)
        {
            ASSERT(!errno);
            errno = EINTR;
            rv = -1;
        }
    }
#ifdef HAVE_WINSOCK2_H
    if (rv < 0)
    {
        errno = WinSockErr2errno(WSAGetLastError());
    }
#endif
    return rv;
}

int tl_selectw(int sock, long int timeout)
{
    fd_set writefds, exceptfds;
    struct timeval tout, *toutptr = NULL;
    int rv;

    /* [i_a] */
    FD_ZERO(&writefds);
    FD_ZERO(&exceptfds);

    /*
     * FD_SET will crash if you pass it sock = -1
     */
    ASSERT(sock_maybe_in_fdset(&writefds, sock));
    if (!sock_maybe_in_fdset(&writefds, sock))
    {
        errno = EINVAL;
        xvperror(DBGvars(NULL), "tl_selectw:FD_SETSIZE (sock = %d)", sock);
        return -1;
    }

    FD_SET(sock, &writefds);
    FD_SET(sock, &exceptfds);

#ifdef HAVE_WINSOCK2_H
    WSASetLastError(0);
#endif
    errno = 0;
    ASSERT(timeout >= 0);
    if (timeout > 0)
    {
        tout.tv_sec = timeout / 1000;
        tout.tv_usec = timeout % 1000;
        toutptr = &tout;
        rv = select(sock + 1, NULL, &writefds, &exceptfds, toutptr);
    }
    else
    {
        /* we could've used pselect() here, but not all systems support it [properly] */
        do
        {
            /* see comment below: must be re-initialized during EVERY round! */
            FD_ZERO(&writefds);
            FD_ZERO(&exceptfds);
            FD_SET(sock, &writefds);
            FD_SET(sock, &exceptfds);

            tout.tv_sec = 1;
            tout.tv_usec = 0;
            toutptr = &tout;
            rv = select(sock + 1, NULL, &writefds, &exceptfds, toutptr);
            DEBUG_DEVEL(DBGvars("select(sock = %d) --> write: %d / exception: %d / return: %d\n"), sock, (int)FD_ISSET(sock, &writefds),
                        (int)FD_ISSET(sock, &exceptfds), rv);
        } while ((rv == 0) && !cfg.stop && !cfg.rbreak);
        /* unlimited time: treat 'timeout' as an error as it can only happen due to stop/break. */
        if (rv == 0)
        {
            ASSERT(!errno);
            errno = EINTR;
            rv = -1;
        }
    }
#ifdef HAVE_WINSOCK2_H
    if (rv < 0)
    {
        errno = WinSockErr2errno(WSAGetLastError());
    }
#endif
    return rv;
}


ssize_t tl_unix_read(int sock, char *buf, size_t len, void *ssl_con)
{
    ssize_t rv = 0;
    long timeout = cfg.read_timeout;
    ssize_t rlen = 0;

    do
    {
#ifdef USE_SSL
        if (!ssl_con || (ssl_con && (my_ssl_data_pending(ssl_con) == 0)))
#endif
        {
#if defined I_FACE && !defined HAVE_MT
            if (cfg.xi_face && cfg.done)
            {
                rv = gui_wait_io(sock, TRUE, timeout);
            }
            else
#endif
            {
                DEBUG_DEVEL(DBGvars("going to wait for read/close event on sock = %d (timeout = %ld)\n"), sock, timeout);
                while ((rv = tl_selectr(sock, timeout)) == -1 && errno == EINTR)
                {
                    DEBUG_DEVEL(DBGvars("waiting for read/close event on sock = %d\n"), sock);

                    if (cfg.rbreak)
                    {
                        DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while waiting for read/close event) -", cfg.stop,
                                    cfg.rbreak);
                        errno = EINTR;
                        cfg.stop = TRUE;
                        /* cfg.rbreak = TRUE; */
                        return -1;
                    }
                    /* in case of 'cfg.stop': try to wind down in orderly fashion, so wait out the timeout. */
                    if (timeout <= 0 && cfg.stop)
                    {
                        DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while waiting for read/close event) -", cfg.stop,
                                    cfg.rbreak);
                        errno = EINTR;
                        /* cfg.rbreak = TRUE; */
                        return -1;
                    }
                }
                if (rv == 0 && timeout > 0)
                {
                    errno = ETIMEDOUT;
                    DEBUG_NET("I/O read sock = %d, timeout = %ld, errno = %d:%s(ETIMEDOUT)\n", sock, timeout, errno, strerror(errno));
                    return -1;
                }
            }

            if (rv < 0)
            {
                DEBUG_NET("I/O read sock = %d, ret = %ld, errno = %d:%s\n", sock, (long)rv, errno, strerror(errno));
                return -1;
            }
        }

        if (cfg.rbreak)
        {
            DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while attempting to read) -", cfg.stop, cfg.rbreak);
            errno = EINTR;
            cfg.stop = TRUE;
            return -1;
        }
        /* in case of 'cfg.stop': try to wind down in orderly fashion, so wait out the timeout. */
        if (timeout <= 0 && cfg.stop)
        {
            DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while attempting to read) -", cfg.stop, cfg.rbreak);
            errno = EINTR;
            /* cfg.rbreak = TRUE; */
            return -1;
        }

#ifdef USE_SSL
        if (ssl_con)
        {
            rv = my_ssl_read(ssl_con, buf, len);
        }
        else
#endif
        {
#ifdef HAVE_WINSOCK2_H
            WSASetLastError(0);
#endif
            errno = 0;          /* [i_a] reset errno or the code will sometimes fail */
#if defined __BEOS__
            /* [i_a] BeOS patch changed; VERY similar to Win32 needs BTW */
            rv = recv(sock, buf, (int)len, 0);
#elif defined(WIN32) || defined(__WIN32)
            WSASetLastError(0);
            rv = recv(sock, buf, (int)len, 0);
            errno = WinSockErr2errno(WSAGetLastError());
#else
            rv = read(sock, buf, len);
#endif
            if (!rv || errno)
            {
                if (errno == EAGAIN)
                {
                    DEBUG_NET("I/O read sock = %d, errno = %d:%s(EAGAIN)\n", sock, errno, strerror(errno));
                }
                else if (errno == EWOULDBLOCK)
                {
                    DEBUG_NET("I/O read sock = %d, errno = %d:%s(EWOULDBLOCK)\n", sock, errno, strerror(errno));
                }
                else if (errno != 0)
                {
                    DEBUG_NET("I/O read sock = %d, ret = %ld / len = %ld, errno = %d:%s\n", sock, (long)rv, (long)len, errno, strerror(errno));
                    return -1;
                }
            }
        }
        if (rv > 0)
        {
            rlen += rv;
            len -= rv;
            buf += rv;
        }

        DEBUG_NET("I/O read sock = %d, ret(len just read) = %ld / len(to go) = %ld / rlen(done) = %ld, errno = %d:%s\n",
                  sock, (long)rv, (long)len, (long)rlen, errno, strerror(errno));

        /* it's 'good enough' when we've fetched a few bytes */
        if (rv == 0 && rlen > 0)
        {
            break;
        }
    } while (rv >= 0 && /* len >= 0 && */ (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK));

    return rlen;
}




/* assumes a non-blocking socket is used! */
ssize_t tl_unix_nbread(int sock, char *buf, size_t len, void *ssl_con)
{
    ssize_t rv = 0;
    ssize_t rlen = 0;

#ifdef USE_SSL
    if (ssl_con && (my_ssl_data_pending(ssl_con) == 0))
    {
        return 0;
    }
#endif

    if (cfg.rbreak)
    {
        DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while attempting to read nonblocking) -", cfg.stop, cfg.rbreak);
        errno = EINTR;
        cfg.stop = TRUE;
        return -1;
    }
    /* in case of 'cfg.stop': try to wind down in orderly fashion, so do the nonblocking read. */
    if (cfg.read_timeout <= 0 && cfg.stop)
    {
        DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while attempting to read nonblocking) -", cfg.stop, cfg.rbreak);
        errno = EINTR;
        /* cfg.rbreak = TRUE; */
        return -1;
    }

#ifdef USE_SSL
    if (ssl_con)
    {
        rv = my_ssl_read(ssl_con, buf, len);
    }
    else
#endif
    {
#ifdef HAVE_WINSOCK2_H
        WSASetLastError(0);
#endif
        errno = 0;              /* [i_a] reset errno or the code will sometimes fail */
#if defined __BEOS__
        /* [i_a] BeOS patch changed; VERY similar to Win32 needs BTW */
        rv = recv(sock, buf, (int)len, 0);
#elif defined(WIN32) || defined(__WIN32)
        WSASetLastError(0);
        rv = recv(sock, buf, (int)len, 0);
        errno = WinSockErr2errno(WSAGetLastError());
#else
        rv = read(sock, buf, len);
#endif
        if (!rv || errno)
        {
            if (errno == EAGAIN)
            {
                DEBUG_NET("I/O nonblocking read sock = %d, errno = %d:%s(EAGAIN)\n", sock, errno, strerror(errno));
                return 0;
            }
            else if (errno == EWOULDBLOCK)
            {
                DEBUG_NET("I/O nonblocking read sock = %d, errno = %d:%s(EWOULDBLOCK)\n", sock, errno, strerror(errno));
                return 0;
            }
            else if (errno != 0)
            {
                DEBUG_NET("I/O nonblocking read sock = %d, ret = %ld / len = %ld, errno = %d:%s\n", sock, (long)rv, (long)len, errno,
                          strerror(errno));
                return -1;
            }
        }
    }
    if (rv > 0)
    {
        rlen += rv;
        len -= rv;
        buf += rv;
    }
    DEBUG_NET("I/O nonblocking read sock = %d, ret(len just read) = %ld / len(to go) = %ld / rlen(done) = %ld, errno = %d:%s\n",
              sock, (long)rv, (long)len, (long)rlen, errno, strerror(errno));

    return rlen;
}




ssize_t tl_unix_write(int sock, const char *buf, size_t len, void *ssl_con)
{
    ssize_t rv = 0;
    ssize_t wlen = 0;
    long timeout = cfg.write_timeout;

    do
    {
#if defined I_FACE && !defined HAVE_MT
        if (cfg.xi_face && cfg.done)
        {
            rv = gui_wait_io(sock, FALSE, timeout);
        }
        else
#endif
        {
            DEBUG_DEVEL(DBGvars("going to wait for write event on sock = %d (timeout = %ld)\n"), sock, timeout);
            while ((rv = tl_selectw(sock, timeout)) == -1 && errno == EINTR)
            {
                DEBUG_DEVEL(DBGvars("waiting for write event on sock = %d\n"), sock);
                if (cfg.rbreak)
                {
                    DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while waiting for write event) -", cfg.stop, cfg.rbreak);
                    errno = EINTR;
                    cfg.stop = TRUE;
                    return -1;
                }
                /* in case of 'cfg.stop': try to wind down in orderly fashion, so wait out the timeout. */
                if (timeout <= 0 && cfg.stop)
                {
                    DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while waiting for write event) -", cfg.stop, cfg.rbreak);
                    errno = EINTR;
                    /* cfg.rbreak = TRUE; */
                    return -1;
                }
            }
            if (rv == 0 && timeout > 0)
            {
                DEBUG_NET("I/O write sock = %d, ret = %ld, errno = %d:%s(ETIMEDOUT)\n", sock, (long)rv, errno, strerror(errno));
                errno = ETIMEDOUT;
                return -1;
            }
        }

        if (rv < 0)
        {
            DEBUG_NET("I/O write sock = %d, ret = %ld, errno = %d:%s\n", sock, (long)rv, errno, strerror(errno));
            ASSERT(errno != 0);
            if (!errno)
                errno = EIO;
            return -1;
        }

        if (cfg.rbreak)
        {
            errno = EINTR;
            DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while attempting to write) -", cfg.stop, cfg.rbreak);
            cfg.stop = TRUE;
            return -1;
        }
        /* in case of 'cfg.stop': try to wind down in orderly fashion, so wait out the timeout. */
        if (timeout <= 0 && cfg.stop)
        {
            DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "I/O (while attempting to write) -", cfg.stop, cfg.rbreak);
            errno = EINTR;
            /* cfg.rbreak = TRUE; */
            return -1;
        }

#ifdef USE_SSL
        if (ssl_con)
        {
            rv = my_ssl_write(ssl_con, buf, len);
        }
        else
#endif
        {
            errno = 0;          /* [i_a] reset errno or the code will sometimes fail */
#if defined __BEOS__
            /* [i_a] BeOS patch changed; VERY similar to Win32 needs BTW */
            rv = send(sock, buf, (int)len, 0);
#elif defined(WIN32) || defined(__WIN32)
            WSASetLastError(0);
            rv = send(sock, buf, (int)len, 0);
            errno = WinSockErr2errno(WSAGetLastError());
#else
            rv = write(sock, buf, len);
#endif
            if (errno != 0 || rv < 0)
            {
                DEBUG_NET("I/O write sock = %d, ret(len just written) = %ld / buf: %p, len(had to write) = %ld, errno = %d:%s\n",
                          sock, (long)rv, buf, (long)len, errno, strerror(errno));
                ASSERT(errno != 0);
                if (!errno)
                    errno = EIO;
                return -1;
            }
        }
        if (rv > 0)
        {
            wlen += rv;
            buf += rv;
            len -= rv;
        }

        DEBUG_NET("I/O write sock = %d, ret(len just written) = %ld / len(to go) = %ld / wlen(done) = %ld, errno = %d:%s\n",
                  sock, (long)rv, (long)len, (long)wlen, errno, strerror(errno));
    } while (rv >= 0 && len > 0);

    return wlen;
}







/* [i_a] BeOS patch changed; VERY similar to Win32 needs BTW */

