/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "bufio.h"
#include "tools.h"
#include "absio.h"
#include "myssl.h"


/* WARNING: must be a power of two! */
#define BUFIO_ALLOC_SECTORSIZE          512


#define BUFIO_ROUND2ALLOC_SECTORSIZE(s)                         \
    (((BUFIO_ALLOC_SECTORSIZE - 1) + (s)) & ~(BUFIO_ALLOC_SECTORSIZE - 1))



static int _bufio_flush(bufio *desc, int override);
static off_t _bufio_seek(bufio *desc, off_t pos, int mode);
static ssize_t _bufio_write(bufio *desc, const char *buf, size_t size);

static int _bufio_tl_write(bufio *desc, const char *buf, size_t size)
{
    int retv;

    if (bufio_is_sock(desc))
    {
        retv = tl_sock_write(bufio_getfd(desc), buf, size, bufio_get_ssl_hook_data(desc));
    }
    else
    {
        retv = tl_file_write(bufio_getfd(desc), buf, size);
    }
    return retv;
}

static int _bufio_tl_read(bufio *desc, char *buf, size_t size)
{
    int retv;

    if (bufio_is_sock(desc))
    {
        retv = tl_sock_read(bufio_getfd(desc), buf, size, bufio_get_ssl_hook_data(desc));
    }
    else
    {
        retv = tl_file_read(bufio_getfd(desc), buf, size);
    }

    return retv;
}

/*
 * will NOT [un?]write buffer content to channel.
 *
 * WILL reallocate the buffer to make sure 'leave_space' bytes
 * are available, at least if you flush the buffer.
 *
 * shift_content is a special hack: it means: leave space
 * BEFORE the current fill, instead of after (following) it.
 *
 * returns !0 if you'll need a bufio_flush to get what you
 * want (shift_content/leave_space)
 */
static int _bufio_fixup_buf(bufio *desc, ssize_t shift_content, ssize_t leave_space)
{
    int retv = 0;

    ASSERT(desc);
    ASSERT(desc->buf_start <= desc->buf_end);
    ASSERT(desc->buf_start <= desc->buf_size);
    ASSERT(desc->buf_end <= desc->buf_size);
    ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

    if (desc->buf_start == desc->buf_end)
    {
        desc->buf_start = shift_content;
        desc->buf_end = shift_content;

        /* all bytes previously written to disc? */
        desc->flags &= ~BUFIO_F_WRBUF_DIRTY;
    }
    else if (((shift_content > 0 && shift_content > desc->buf_start)
             || (leave_space > 0 && desc->buf_end + leave_space > desc->buf_size))
            && (shift_content + desc->buf_end - desc->buf_start + leave_space <= desc->buf_size))
    {
        /*
         * move data if shift_content bytes can fit before it as it is (but not now)
         *
         *     OR
         *
         * move data if we need at least 'leave_space' bytes of extra space for our subsequent actions
         * and there's room for that as it is (but not now)
         */
        ASSERT(desc->buf_start < desc->buf_end);
        ASSERT(desc->buf_end < desc->buf_size);
        memmove(desc->buf + shift_content, desc->buf + desc->buf_start, desc->buf_end - desc->buf_start);
        desc->buf_end -= desc->buf_start - shift_content;
        desc->buf_start = shift_content;
        ASSERT(desc->buf_start < desc->buf_end);
        ASSERT(desc->buf_start < desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);
    }

    /*
     * By now the real content of the buffer has been moved into an optimal position
     * if it could be done without resizing the buffer.
     *
     * So what is left now is to see how much we need to resize the buffer
     * and then move the real data to the proper position within the new
     * and resized buffer.
     *
     * make sure there's EXACTLY enough space BEFORE the content: 'shift_content'.
     */
    if (shift_content + desc->buf_end - desc->buf_start + leave_space > desc->buf_size)
    {
        ASSERT(desc->buf_start <= desc->buf_end);
        ASSERT(desc->buf_start <= desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);
        /*
         * The fun bit is that a resized buffer should NOT assume a successful flush of the currently
         * carried data. Of course, we don't HAVE to assume that, but it is just a wee bit safer
         * going when a subsequent flush() fails to perform after this call. Being overly cautious I am. ;-)
         */
        desc->buf_size = BUFIO_ROUND2ALLOC_SECTORSIZE(shift_content + desc->buf_end - desc->buf_start + leave_space);
        desc->buf = _realloc(desc->buf, desc->buf_size);
        ASSERT(desc->buf_start <= desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);
        /*
         * now move the current valid content to
         * the 'optimal' position (and the only
         * valid one after we leave) this call.
         */
        if (desc->buf_start != shift_content && desc->buf_start != desc->buf_end)
        {
            ASSERT(desc->buf_start < desc->buf_end);
            memmove(desc->buf + shift_content, desc->buf + desc->buf_start, desc->buf_end - desc->buf_start);
            desc->buf_end -= desc->buf_start - shift_content;
            desc->buf_start = shift_content;
            ASSERT(desc->buf_start < desc->buf_end);
            ASSERT(desc->buf_start < desc->buf_size);
            ASSERT(desc->buf_end <= desc->buf_size);
        }
        else
        {
            ASSERT(desc->buf_start == shift_content);
            ASSERT(desc->buf_end >= shift_content);
        }
    }

    retv = (desc->buf_end + leave_space > desc->buf_size);
    return retv;
}

static int _bufio_flock(bufio *desc, bool_t lock_mode)
{
    int retv;

    ASSERT(desc);
    ASSERT(bufio_getfd(desc) >= 0);
    retv = _flock(bufio_getfd(desc), desc->filename, desc->open_flags, lock_mode);

    return retv;
}

static int _bufio_funlock(bufio *desc)
{
    int retv;

    ASSERT(desc);
    ASSERT(bufio_getfd(desc) >= 0);
    retv = _bufio_flush(desc, TRUE);
    retv |= _funlock(bufio_getfd(desc));

    return retv;
}

static void _bufio_mutex_lock(bufio *desc)
{
    ASSERT(desc);

#ifdef HAVE_MT
    if (desc->mutex)
    {
        pthread_mutex_lock(desc->mutex);
    }
#endif
}

static void _bufio_mutex_unlock(bufio *desc)
{
    ASSERT(desc);

#ifdef HAVE_MT
    if (desc->mutex)
    {
        pthread_mutex_unlock(desc->mutex);
    }
#endif
}

static size_t _bufio_set_minbuf(bufio *desc, size_t size)
{
    ASSERT(desc);
    size = ((desc->flags & BUFIO_F_IS_LOGFILE) ? TL_MAX(size, BUFIO_LOG_WRITEBUFSIZE) : TL_MAX(size, BUFIO_BUF_SIZE));

    if (size > desc->buf_size || !desc->buf)
    {
        ASSERT(size > 0);
        if (!desc->buf)
        {
            desc->buf = _malloc(BUFIO_ROUND2ALLOC_SECTORSIZE(size));
        }
        else
        {
            desc->buf = _realloc(desc->buf, BUFIO_ROUND2ALLOC_SECTORSIZE(size));
        }
        desc->buf_size = size;
    }

    return desc->buf_size;
}

static void _bufio_reset(bufio *desc)
{
    ASSERT(desc);
    ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

    desc->buf_start = 0;
    desc->buf_end = 0;
    desc->flags &= ~BUFIO_F_WRBUF_DIRTY;
}

static int _bufio_truncate(bufio *desc, off_t len)
{
    int retv = 0;

    ASSERT(desc);
    ASSERT(bufio_is_open(desc));
//printf("(%s:%d)\n", __FILE__, __LINE__);
#if 0
    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_RD_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return -1;
    }
#endif

    if (!bufio_is_sock(desc))
    {
        retv = _bufio_flush(desc, TRUE);
//printf("(%s:%d)\n", __FILE__, __LINE__);
        if (!retv)
        {
            _bufio_reset(desc); /* MANDATORY to make sure the buffer is discarded: it is invalid after a file pointer move anyhow! */

//printf("(%s:%d)\n", __FILE__, __LINE__);
            retv = ftruncate(bufio_getfd(desc), len);
//printf("(%s:%d)\n", __FILE__, __LINE__);
        }
//printf("(%s:%d)\n", __FILE__, __LINE__);
    }

    return retv;
}

static int _bufio_set_flags(bufio *desc, int flags, int mask)
{
    int retv;

    ASSERT(desc);

    desc->flags &= ~mask;
    desc->flags |= flags;
    retv = desc->flags;

    return retv;
}


/*
 * FIXME: interface of this function is a mess. Really.
 */
static int _bufio_open2spec(bufio *desc, int open_flags, int open_mask)
{
    int retv;

    ASSERT(desc);
    DEBUG_BUFIO("open2spec(%d, %s)\n", desc->fd, mk_native(desc->filename));
//printf("(%s:%d)\n", __FILE__, __LINE__);
    if (!bufio_is_open(desc))
    {
        ASSERT(desc->fd < 0);
        if (desc->flags & BUFIO_F_APPEND)
        {
            open_flags &= ~O_TRUNC;
            open_flags |= O_APPEND;
        }
        else
        {
            open_flags &= ~O_APPEND;
            open_flags |= O_TRUNC;
        }
//printf("(%s:%d)\n", __FILE__, __LINE__);
        if (!open_mask)
        {
            desc->fd = tl_open(desc->filename, open_flags, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
        }
        else
        {
            desc->fd = tl_open(desc->filename, open_flags, open_mask);
        }
//printf("(%s:%d)\n", __FILE__, __LINE__);
        desc->open_flags = open_flags;
        if (desc->fd < 0)
        {
            DEBUG_BUFIO("open2spec(%d, %s) --> %d/%s\n", desc->fd, mk_native(desc->filename), errno, strerror(errno));
            desc->bufio_errno = errno;
        }
//printf("(%s:%d)\n", __FILE__, __LINE__);
    }

//printf("(%s:%d)\n", __FILE__, __LINE__);
    retv = !bufio_is_open(desc);
    if (!retv)
    {
//printf("(%s:%d)\n", __FILE__, __LINE__);
        if (desc->open_flags & (O_WRONLY | O_RDWR))
        {
            _bufio_set_flags(desc, BUFIO_F_DELAY_WRITE, 0);
        }
        _bufio_set_minbuf(desc, desc->buf_size);
//printf("(%s:%d)\n", __FILE__, __LINE__);

        /* retv->fd = fd; */
        /* retv->filename = tl_strdup(filename); */
        desc->flags &= ~BUFIO_F_APPEND;
        desc->flags |= ((open_flags & O_APPEND) ? BUFIO_F_APPEND : 0);
        if (open_flags & O_TRUNC)
        {
//printf("(%s:%d)\n", __FILE__, __LINE__);
            retv = _bufio_truncate(desc, 0);
        }
    }
//printf("(%s:%d)\n", __FILE__, __LINE__);

    return retv;
}


static int _bufio_flush(bufio *desc, int override)
{
    int l = 0;
    int retv = 0;

    ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

//printf("(%s:%d)\n", __FILE__, __LINE__);
    if (desc->flags & BUFIO_F_WRBUF_DIRTY)
    {
        /* flush content, even under the worst conditions. BUT only do this for *FILES*! */
        bool_t st = cfg.stop;
        bool_t br = cfg.rbreak;

//printf("(%s:%d)\n", __FILE__, __LINE__);
        if (bufio_is_sock(desc))
        {
            cfg.stop = FALSE;
            cfg.rbreak = FALSE;
        }
        ASSERT(desc->flags & BUFIO_F_DELAY_WRITE);
        l = desc->buf_end - desc->buf_start;
        ASSERT(l >= 0);
//printf("(%s:%d)\n", __FILE__, __LINE__);
        while (l > 0 && (!override ? !cfg.stop && !cfg.rbreak : TRUE))
        {
            int r = 1;

//printf("(%s:%d)\n", __FILE__, __LINE__);
            errno = 0;
            if (bufio_is_sock(desc))
            {
                DEBUG_DEVEL(DBGvars("going to wait for write event on sock = %d (timeout = %ld)\n"), desc->fd, cfg.write_timeout);
                r = tl_selectw(desc->fd, (int)cfg.write_timeout);
                DEBUG_DEVEL(DBGvars("waited for write event on sock = %d --> %d\n"), desc->fd, r);
            }
            if (r > 0)
            {
//printf("(%s:%d)\n", __FILE__, __LINE__);
                r = _bufio_tl_write(desc, desc->buf + desc->buf_start, l);
//printf("(%s:%d)\n", __FILE__, __LINE__);
                if (r > 0)
                {
                    desc->buf_start += r;
                    ASSERT(desc->buf_start <= desc->buf_end);
                    ASSERT(desc->buf_start <= desc->buf_size);
                    ASSERT(desc->buf_end <= desc->buf_size);
                    l -= r;
                }
                if (errno != 0 || r <= 0)
                {
                    /* failed to write some bytes (or any byte at all) */
                    ASSERT(errno != 0);
                    if (!errno)
                        errno = EIO;
                    desc->bufio_errno = errno;
//printf("(%s:%d)\n", __FILE__, __LINE__);
                    retv = -1;
                    break;
                }
            }
            else
            {
                /* select() timeout or error */
//printf("(%s:%d)\n", __FILE__, __LINE__);
                desc->bufio_errno = errno;
                if (!desc->bufio_errno)
                {
                    errno = desc->bufio_errno = ETIMEDOUT;
                }
//printf("(%s:%d)\n", __FILE__, __LINE__);
                retv = -1;
                break;
            }
        }
        /*
         * cfg.stop and/or cfg.break can be altered asynchronously (signals),
         * so take care about that by not destroying those possible changes here:
         */
        ASSERT(!retv && !cfg.stop && !cfg.rbreak ? (l == 0) : TRUE);
        ASSERT(!retv && !cfg.stop && !cfg.rbreak ? ((desc->buf_end - desc->buf_start) == 0) : TRUE);
        cfg.stop |= st;
        cfg.rbreak |= br;
//printf("cfg.stop = %d, cfg.rbreak = %d -- remind_do() restored cfg.stop/cfg.rbreak", cfg.stop, cfg.rbreak);
        ASSERT(desc->flags & BUFIO_F_WRBUF_DIRTY);
        ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
        ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

//printf("(%s:%d)\n", __FILE__, __LINE__);
        if (_bufio_fixup_buf(desc, 0, 0))
        {
//printf("(%s:%d)\n", __FILE__, __LINE__);
            if (!retv)
            {
                ASSERT(0);
                errno = EIO;
                desc->bufio_errno = EIO;
                retv = -1;
            }
        }
//printf("(%s:%d)\n", __FILE__, __LINE__);
        ASSERT(!retv && !cfg.stop && !cfg.rbreak ? ((desc->buf_end - desc->buf_start) == 0) : TRUE);
        ASSERT(!retv && !cfg.stop && !cfg.rbreak ? !(desc->flags & BUFIO_F_WRBUF_DIRTY) : TRUE);
        ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
        ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    }
    else
    {
//printf("(%s:%d)\n", __FILE__, __LINE__);
        /* ASSERT(desc->buf_end == desc->buf_start); */
        _bufio_fixup_buf(desc, 0, 0);
//printf("(%s:%d)\n", __FILE__, __LINE__);
        ASSERT(!(desc->flags & BUFIO_F_WRBUF_DIRTY));
//printf("(%s:%d)\n", __FILE__, __LINE__);
    }
//printf("(%s:%d)\n", __FILE__, __LINE__);

    if (!retv && desc->buf_end - desc->buf_start == 0)
    {
        desc->flags &= ~BUFIO_F_WRBUF_DIRTY;
    }
//printf("(%s:%d)\n", __FILE__, __LINE__);

    return retv;
}


/*
 * Make sure write data is flushed before we attempt to read into
 * the buffer!
 *
 * In 'append' mode, seek to end when we read data before and
 * are going to write now.
 */
static int _bufio_check_and_fix_last_action(bufio *desc, int to_be_state)
{
    int retv = 0;

    ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

    if (to_be_state & BUFIO_F_RD_WAS_LAST)
    {
        if (desc->flags & BUFIO_F_WR_WAS_LAST)
        {
            if (desc->flags & BUFIO_F_WRBUF_DIRTY)
            {
                retv = _bufio_flush(desc, FALSE);
                if (!retv)
                {
                    ASSERT(desc->buf_end - desc->buf_start == 0);
                    ASSERT(desc->buf_end == 0);
                    /* _bufio_reset(desc); */
                }
            }
            else
            {
//printf("(%s:%d) last read: %d, last write: %d\n", __FILE__, __LINE__, (desc->flags & BUFIO_F_RD_WAS_LAST), (desc->flags & BUFIO_F_WR_WAS_LAST));
                ASSERT(desc->buf_end - desc->buf_start == 0);
            }
        }
    }
    else if (to_be_state & BUFIO_F_WR_WAS_LAST)
    {
        if (desc->flags & BUFIO_F_RD_WAS_LAST)
        {
            /* we are going to write; were reading before: kill buffer! */
            ASSERT(!(desc->flags & BUFIO_F_WRBUF_DIRTY));
            _bufio_reset(desc);

            if (!bufio_is_sock(desc) && (desc->flags & BUFIO_F_APPEND))
            {
                /* move file position to end of file again; do NOT do this for sockets! */
                _bufio_seek(desc, 0, SEEK_END);
                retv = (desc->bufio_errno != 0);
            }
        }
    }
    return retv;
}

static off_t _bufio_seek(bufio *desc, off_t pos, int mode)
{
    off_t retv = 0;

    ASSERT(desc);
    ASSERT(desc->fd >= 0);
    if (!bufio_is_sock(desc))
    {
#if 0
        if (_bufio_check_and_fix_last_action(desc, BUFIO_F_RD_WAS_LAST))
        {
            /* error condition! */
            ASSERT(desc->bufio_errno);
            errno = desc->bufio_errno;
            return -1;
        }
#endif

        retv = _bufio_flush(desc, TRUE);
        if (!retv)
        {
            _bufio_reset(desc); /* MANDATORY to make sure the buffer is discarded: it is invalid after a file pointer move anyhow! */

#if defined(HAVE_LSEEK64)
            retv = lseek64(bufio_getfd(desc), pos, mode);
#elif defined(HAVE_LLSEEK)
            retv = llseek(bufio_getfd(desc), pos, mode);
#else
            retv = lseek(bufio_getfd(desc), pos, mode);
#endif
            if (retv < 0)
            {
                desc->bufio_errno = errno;
                retv = -1;
            }
        }
    }

    return retv;
}


static int _bufio_vprintf(bufio *desc, const char *msg, va_list args)
{
    size_t n;
    char *dst;

    ASSERT(desc);
    ASSERT(desc->fd >= 0);
    ASSERT(desc->buf);
    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_WR_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return 0;
    }

    ASSERT(desc->buf_size >= desc->buf_end);
    n = desc->buf_size - desc->buf_end;
    if (n < TL_MIN(desc->buf_size, BUFIO_ADVISED_WRITELN_BUFSIZE))
    {
        if (_bufio_fixup_buf(desc, 0, TL_MIN(desc->buf_size, BUFIO_ADVISED_WRITELN_BUFSIZE)))
        {
            if (_bufio_flush(desc, FALSE))
            {
                ASSERT(desc->bufio_errno);
            }
        }
        ASSERT(desc->buf_size >= desc->buf_end);
        n = desc->buf_size - desc->buf_end;
    }
    if (n > 1)
    {
        ssize_t l;

        dst = desc->buf + desc->buf_end;
        l = vsnprintf(dst, n, msg, args);
        if (l < 0)
        {
            /* vsnprintf() was truncated... (MSVC et al) -> guestimate 'acceptable' buffer space */
            l = desc->buf_size / 2;
        }
        if (l >= n)
        {
            /* make sure we have enough space to store l chars in bufio: */
            if (_bufio_fixup_buf(desc, 0, TL_MIN((ssize_t)desc->buf_size, l + 1)))
            {
                if (_bufio_flush(desc, FALSE))
                {
                    ASSERT(desc->bufio_errno);
                }
            }
            ASSERT(desc->buf_size >= desc->buf_end);
            n = desc->buf_size - desc->buf_end;
            dst = desc->buf + desc->buf_end;
            l = vsnprintf(dst, n, msg, args);
            ASSERT(l < n);
        }
        dst[n - 1] = 0;
        n = strlen(dst);
        desc->buf_end += n;
        desc->flags |= BUFIO_F_WRBUF_DIRTY;
    }
    else
    {
        n = 0;
    }
    ASSERT(desc->buf_start <= desc->buf_end);
    ASSERT(desc->buf_start <= desc->buf_size);
    ASSERT(desc->buf_end <= desc->buf_size);

    desc->flags &= ~BUFIO_F_RD_WAS_LAST;
    desc->flags |= BUFIO_F_WR_WAS_LAST;

    ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

    if (!(desc->flags & BUFIO_F_DELAY_WRITE))
    {
        if (_bufio_flush(desc, FALSE))
        {
            ASSERT(desc->bufio_errno);
            n = 0;
        }
    }

    return (int)n;
}



/*
 * ===========================================================================
 */

#undef bufio_new
bufio *bufio_new(DBGdecl(bool_t with_mutex), size_t requested_bufsize)
{
    bufio *retv;

    retv = tld_malloc(DBGpass(sizeof(*retv)));
    retv->buf_size = (requested_bufsize == 0 ? BUFIO_BUF_SIZE : requested_bufsize);
    retv->buf = tld_malloc(DBGpass(retv->buf_size));
    retv->fd = -1;
    retv->open_flags = 0;
    retv->buf_start = 0;
    retv->buf_end = 0;
    retv->bufio_errno = 0;
    retv->eof = 0;
    retv->is_sock = 0;
    retv->ssl_hook_data = NULL;
    retv->no_close = FALSE;
    retv->mutex = NULL;
    retv->filename = NULL;
    retv->flags = 0;
#ifdef HAVE_MT
    if (with_mutex)
    {
        retv->mutex = tld_malloc(DBGpass(sizeof(retv->mutex[0])));
        pthread_mutex_init(retv->mutex, NULL);
    }
#endif

    return retv;
}

size_t bufio_set_minbuf(bufio *desc, size_t size)
{
    size_t retv;

    ASSERT(desc);
    _bufio_mutex_lock(desc);

    retv = _bufio_set_minbuf(desc, size);

    _bufio_mutex_unlock(desc);

    return retv;
}

/*
 * Return a fully initialized bufio structure for the given handle.
 *
 * Return NULL if the handle is invalid.
 */
bufio *bufio_sock_fdopen(int fd, int open_flags)
{
    bufio *retv;

    if (fd < 0)
        return NULL;

    retv = bufio_new(DBGvars(FALSE), 0);
    if (retv)
    {
        retv->fd = fd;
        retv->is_sock = 1;
        ASSERT(!(retv->flags & BUFIO_F_DELAY_WRITE));
        if (open_flags & (O_WRONLY | O_RDWR))
        {
            /* _bufio_set_flags(retv, BUFIO_F_DELAY_WRITE, 0); */
        }
        _bufio_set_minbuf(retv, retv->buf_size);

        /* can't 'truncate' socket I/O: */
        open_flags &= ~O_TRUNC;
        open_flags |= O_APPEND;
        retv->flags |= BUFIO_F_APPEND;
        retv->open_flags = open_flags;
    }

    return retv;
}

bufio *bufio_fdopen(int fd, int open_flags)
{
    bufio *retv;

    if (fd < 0)
        return NULL;

    retv = bufio_new(DBGvars(FALSE), 0);
    if (retv)
    {
        retv->fd = fd;
        if (open_flags & (O_WRONLY | O_RDWR))
        {
            _bufio_set_flags(retv, BUFIO_F_DELAY_WRITE, 0);
        }
        _bufio_set_minbuf(retv, retv->buf_size);

        retv->flags &= ~BUFIO_F_APPEND;
        if (open_flags & O_APPEND)
        {
            open_flags &= ~O_TRUNC;
            /* open_flags |= O_APPEND; */
            retv->flags |= BUFIO_F_APPEND;
        }
        else
        {
            if (open_flags & O_TRUNC)
            {
                _bufio_truncate(retv, 0);
            }
            open_flags &= ~O_APPEND;
            /* open_flags |= O_TRUNC; */
        }
        retv->open_flags = open_flags;
    }

    return retv;
}

/*
 * FIXME: interface of this function is a mess. Really.
 */
int bufio_open2spec(bufio *desc, int open_flags, int open_mask)
{
    int retv;

    ASSERT(desc);
    _bufio_mutex_lock(desc);

    retv = _bufio_open2spec(desc, open_flags, open_mask);

    _bufio_mutex_unlock(desc);

    return retv;
}


bufio *bufio_open(const char *filename, int flags)
{
    bufio *desc;

    DEBUG_BUFIO("open(%s)\n", filename);
    desc = bufio_new(DBGvars(FALSE), 0);
    desc->filename = tl_strdup(filename);
    desc->open_flags = flags;
    desc->flags |= ((flags | O_APPEND) ? BUFIO_F_APPEND : 0);
    if (_bufio_open2spec(desc, flags, 0))
    {
        bufio_close(desc);
        desc = NULL;
    }

    return desc;
}

bufio *bufio_copen(const char *filename, int flags, int mask)
{
    bufio *desc;

    desc = bufio_new(DBGvars(FALSE), 0);
    desc->filename = tl_strdup(filename);
    desc->open_flags = flags;
    desc->flags |= ((flags & O_APPEND) ? BUFIO_F_APPEND : 0);
    if (_bufio_open2spec(desc, flags, mask))
    {
        bufio_close(desc);
        desc = NULL;
    }

    return desc;
}

bufio *bufio_new_sslcon(bufio *sock, void *con)
{
    bufio *desc;

    if (!sock)
        return NULL;

    desc = bufio_new(DBGvars(!!sock->mutex), sock->buf_size);

    _bufio_mutex_lock(desc);

    desc->ssl_hook_data = con;
    desc->is_sock = 1;
    ASSERT(!(desc->flags & BUFIO_F_DELAY_WRITE));
    desc->fd = bufio_getfd(sock);
    /*
     * FIXME
     * WARNING: this copying of file handles is DANGEROUS!
     * I'm trying to fix it here, after looking at the usage of this function:
     * NOTE that I 'dont_close' flag the ORIGINAL bufio, instead of the newly
     * generated one as you'd normally expect.
     */
    sock->no_close = TRUE;

    _bufio_mutex_unlock(desc);

    return desc;
}

void bufio_free(bufio *desc)
{
    if (!desc)
        return;

    ASSERT(!desc->bufio_errno && !cfg.stop && !cfg.rbreak ? !(desc->flags & BUFIO_F_WRBUF_DIRTY) : TRUE);
#ifdef HAVE_MT
    if (desc->mutex && pthread_mutex_destroy(desc->mutex))
    {
        xperror("mt_semaphore_destroy - bufio_free");
    }
    _free(desc->mutex);
#else
    ASSERT(!desc->mutex);
#endif
    _free(desc->filename);
    _free(desc->buf);
    _free(desc);
}

int bufio_close(bufio *desc)
{
    int retv = 0;

    if (!desc)
        return 0;

#if 0
    if (td > 0 && td == desc->fd)
    {
        printf("hello!\n");
    }
#endif

    _bufio_mutex_lock(desc);

    if (desc->flags & BUFIO_F_WRBUF_DIRTY)
    {
        ASSERT(desc->flags & BUFIO_F_WRBUF_DIRTY);
        ASSERT(desc->flags & BUFIO_F_DELAY_WRITE);
        ASSERT(desc->flags & BUFIO_F_WR_WAS_LAST);
        if (_bufio_flush(desc, bufio_is_open_file(desc)))
        {
            ASSERT(desc->bufio_errno);
            retv = -1;
        }
        else
        {
            ASSERT(!cfg.stop && !cfg.rbreak ? !(desc->flags & BUFIO_F_WRBUF_DIRTY) : TRUE);
        }
    }
    ASSERT(!desc->bufio_errno && !cfg.stop && !cfg.rbreak ? !(desc->flags & BUFIO_F_WRBUF_DIRTY) : TRUE);

#ifdef USE_SSL
    if (bufio_get_ssl_hook_data(desc))
    {
        my_ssl_connection_destroy(bufio_get_ssl_hook_data(desc));
        desc->ssl_hook_data = NULL;
        desc->fd = -1;
    }
#endif

    if (desc->fd >= 0 && !desc->no_close)
    {
        if (bufio_is_sock(desc))
        {
            shutdown(desc->fd, SHUT_RDWR);
#ifdef __BEOS__
            retv |= closesocket(desc->fd);
#elif defined(WIN32) || defined(__WIN32)
            retv |= closesocket(desc->fd);
#else
            retv |= close(desc->fd);
#endif
        }
        else
        {
            retv |= close(desc->fd);
        }
    }

    desc->fd = -1;

    _bufio_mutex_unlock(desc);

    bufio_free(desc);

    return retv;
}

int bufio_flush(bufio *desc)
{
    int retv;

    _bufio_mutex_lock(desc);

    retv = _bufio_flush(desc, bufio_is_open_file(desc));

    _bufio_mutex_unlock(desc);

    return retv;
}


int bufio_truncate(bufio *desc, off_t len)
{
    int retv;

    ASSERT(desc);
    ASSERT(bufio_is_open(desc));

    _bufio_mutex_lock(desc);

    retv = _bufio_truncate(desc, len);

    _bufio_mutex_unlock(desc);

    return retv;
}

int bufio_init_mutex(bufio *desc, bool_t has_a_mutex)
{
    int retv = 0;

    ASSERT(desc);
#ifdef HAVE_MT
    if (!has_a_mutex)
    {
        if (desc->mutex)
        {
            if ((retv = pthread_mutex_destroy(desc->mutex)) != 0)
            {
                xperror("mt_semaphore_destroy - bufio_init_mutex");
            }
            _free(desc->mutex);
            ASSERT(!desc->mutex);
        }
    }
    else
    {
        if (!desc->mutex)
        {
            desc->mutex = _malloc(sizeof(desc->mutex[0]));
            retv = pthread_mutex_init(desc->mutex, NULL);
        }
    }
#endif

    return retv;
}

int bufio_set_flags(bufio *desc, int flags, int mask)
{
    int retv;

    ASSERT(desc);
    _bufio_mutex_lock(desc);

    retv = _bufio_set_flags(desc, flags, mask);

    _bufio_mutex_unlock(desc);

    return retv;
}

int bufio_flock(bufio *desc, bool_t lock_mode)
{
    int retv;

    ASSERT(desc);
    ASSERT(desc->fd >= 0);

    _bufio_mutex_lock(desc);

    retv = _bufio_flock(desc, lock_mode);

    _bufio_mutex_unlock(desc);

    return retv;
}

int bufio_funlock(bufio *desc)
{
    int retv;

    ASSERT(desc);
    ASSERT(desc->fd >= 0);

    _bufio_mutex_lock(desc);

    retv = _bufio_funlock(desc);

    _bufio_mutex_unlock(desc);

    return retv;
}

/*
 * lock using mutex;
 * use _flock() ONLY once we have to write/flush the buffer to disc!
 *
 * return !0 if non_blocking detected a previous lock.
 *
 * When succesful, *nonblocking_dst_desc will have been set to a
 * image copy of your IO descriptor 'desc', but with the mutex temporarily DISabled
 * until you 'recover' by calling bufio_release(). See the notes below:
 *
 * Note to prevent deadlocks due to using calls like bufio_puts() et al
 * within a bufio_lock() ... bufio_release() section, bufio_lock() provides you
 * with a direct copy of the IO descriptor 'desc', but without any
 * mutex in there, so bufio_puts() will not try to lock you out.
 *
 * Note however that you MUST use bufio_release() to 'recover' your original
 * IO descriptor 'desc' to the proper state!
 */
int bufio_lock(bufio *desc, bool_t non_blocking, bufio **nonlocking_dst_desc)
{
    int retv = 0;

    if (!nonlocking_dst_desc)
    {
        desc->bufio_errno = EINVAL;
        return -1;
    }
    *nonlocking_dst_desc = NULL;

    ASSERT(desc);
#if 0
    ASSERT(desc->fd >= 0);
#endif
#ifdef HAVE_MT
    if (desc->mutex)
    {
        if (non_blocking)
        {
            retv = pthread_mutex_trylock(desc->mutex);
        }
        else
        {
            retv = pthread_mutex_lock(desc->mutex);
        }
    }
#endif

    if (!retv)
    {
        nonlocking_dst_desc[0] = _malloc(sizeof(nonlocking_dst_desc[0][0]));
        memcpy(nonlocking_dst_desc[0], desc, sizeof(nonlocking_dst_desc[0][0])); /* copy values across */
#ifdef HAVE_MT
        nonlocking_dst_desc[0]->mutex = NULL;   /* but ditch the mutex for now... */
#endif
    }
    return retv;
}

/*
 * Note: a call to this function ASSUMES you had a previous SUCCESFUL call
 * to bufio_lock().
 *
 * It also ASSUMES you have only really _used_ the 'nonlocking_desc'
 * descriptor since then, as it will OVERWRITE any changes you made to the
 * 'desc' descriptor due to using it in other bufio_* IO operations.
 */
int bufio_release(bufio *desc, bufio **nonlocking_desc)
{
    int retv = 0;

    ASSERT(desc);
    if (!nonlocking_desc)
    {
        desc->bufio_errno = EINVAL;
        return -1;
    }
    if (*nonlocking_desc)
    {
#ifdef HAVE_MT
        pthread_mutex_t *mutex = desc->mutex;   /* remember the mutex! */
#endif
        memcpy(desc, nonlocking_desc[0], sizeof(bufio)); /* copy values BACK! */
#ifdef HAVE_MT
        /* but restore the original mutex! */
        desc->mutex = mutex;
#endif
        _free(nonlocking_desc[0]);
        ASSERT(nonlocking_desc[0] == NULL);
    }
#if 0
    ASSERT(desc->fd >= 0);
#endif
#ifdef HAVE_MT
    if (desc->mutex)
    {
        retv = pthread_mutex_unlock(desc->mutex);
    }
#endif

    return retv;
}

off_t bufio_seek(bufio *desc, off_t pos, int mode)
{
    off_t retv;

    ASSERT(desc);
    ASSERT(desc->fd >= 0);

    _bufio_mutex_lock(desc);

    retv = _bufio_seek(desc, pos, mode);

    _bufio_mutex_unlock(desc);

    return retv;
}

int bufio_printf(bufio *desc, const char *msg, ...)
{
    int retv;
    va_list args;

    va_start(args, msg);
    ASSERT(desc);
    ASSERT(desc->fd >= 0);
    ASSERT(desc->buf);

    _bufio_mutex_lock(desc);

    retv = _bufio_vprintf(desc, msg, args);

    _bufio_mutex_unlock(desc);

    va_end(args);

    return retv;
}


int bufio_vprintf(bufio *desc, const char *msg, va_list args)
{
    int retv;

    ASSERT(desc);
    ASSERT(desc->fd >= 0);
    ASSERT(desc->buf);

    _bufio_mutex_lock(desc);

    retv = _bufio_vprintf(desc, msg, args);

    _bufio_mutex_unlock(desc);

    return retv;
}

int bufio_puts(bufio *desc, const char *str)
{
    return bufio_write(desc, str, strlen(str));
}

int bufio_puts_m(bufio *desc, const char *str, ...)
{
    int retv;

    va_list args;

    va_start(args, str);

    _bufio_mutex_lock(desc);

    retv = _bufio_write(desc, str, strlen(str));
    for (; retv >= 0;)
    {
        int rv;

        str = va_arg(args, const char *);
        if (str == NULL)
            break;
        rv = _bufio_write(desc, str, strlen(str));
        if (rv < 0)
        {
            retv = rv;          /* break; */
        }
        else
        {
            retv += rv;
        }
    }

    _bufio_mutex_unlock(desc);

    va_end(args);

    return retv;
}






size_t bufio_asciidump(bufio *desc, const char *cbuf, size_t csize, const char *title)
{
    size_t n;
    char *dst;
    size_t dstsize = csize * 5 + 81;
    int fx = 0;

    _bufio_mutex_lock(desc);

    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_WR_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return 0;
    }
    if (title)
    {
        fx = _bufio_write(desc, title, strlen(title));
    }
    /* ASSERT(desc->flags & BUFIO_F_DELAY_WRITE); */
    if (fx >= 0)
    {
        fx = _bufio_fixup_buf(desc, 0, dstsize);
        if (fx)
        {
            fx = 0;
            if (_bufio_flush(desc, FALSE))
            {
                ASSERT(desc->bufio_errno);
                fx = 1;
            }
        }
        if (!fx)
        {
            ASSERT(desc->buf_size >= desc->buf_end);
            n = desc->buf_size - desc->buf_end;
            if (n < dstsize)
            {
                _bufio_set_minbuf(desc, dstsize + desc->buf_end /* - desc->buf_start */);
                n = desc->buf_size - desc->buf_end;
                ASSERT(n >= dstsize);
            }
            dst = desc->buf + desc->buf_end;
            tl_asciidump_content(dst, n, cbuf, csize);
            n = strlen(dst);
            desc->buf_end += n;
            desc->flags |= BUFIO_F_WRBUF_DIRTY;
        }
        else
        {
            n = 0;
        }
        ASSERT(desc->buf_start <= desc->buf_end);
        ASSERT(desc->buf_start <= desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);
    }
    else
    {
        ASSERT(desc->bufio_errno);
        n = 0;
    }
    desc->flags &= ~BUFIO_F_RD_WAS_LAST;
    desc->flags |= BUFIO_F_WR_WAS_LAST;

    ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

    if (!(desc->flags & BUFIO_F_DELAY_WRITE))
    {
        if (_bufio_flush(desc, FALSE))
        {
            ASSERT(desc->bufio_errno);
            n = 0;
        }
    }

    _bufio_mutex_unlock(desc);

    return n;
}


ssize_t bufio_write(bufio *desc, const char *buf, size_t size)
{
    ssize_t retv;

    _bufio_mutex_lock(desc);

    retv = _bufio_write(desc, buf, size);

    _bufio_mutex_unlock(desc);

    return retv;
}

/*
 * specify a series of byte blocks to write,
 * terminate with a NULL pointer.
 */
ssize_t bufio_write_m(bufio *desc, const char *buf, size_t size, ...)
{
    ssize_t len;
    va_list args;

    va_start(args, size);

    _bufio_mutex_lock(desc);

    len = _bufio_write(desc, buf, size);

    for (; len >= 0;)
    {
        size_t s;
        ssize_t l;

        buf = va_arg(args, const char *);
        if (buf == NULL)
            break;
        s = va_arg(args, size_t);
        l = _bufio_write(desc, buf, s);
        if (l < 0)
        {
            len = l;            /* break; */
        }
        else
        {
            len += l;
        }
    }

    _bufio_mutex_unlock(desc);

    va_end(args);
    return len;
}

static ssize_t _bufio_write(bufio *desc, const char *buf, size_t size)
{
    ssize_t n = 0;
    char *dst;
    int fx;

//printf("(%s:%d) _bufio_write(): flags & delay_write = %d\n", __FILE__, __LINE__, (desc->flags & BUFIO_F_DELAY_WRITE));
    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_WR_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return -1;
    }

    if (desc->flags & BUFIO_F_DELAY_WRITE)
    {
        fx = _bufio_fixup_buf(desc, 0, size);
        if (fx)
        {
            fx = 0;
            if (_bufio_flush(desc, FALSE))
            {
                ASSERT(desc->bufio_errno);
                fx = 1;
            }
        }
        if (!fx)
        {
            ASSERT(desc->buf_size >= desc->buf_end);
            n = desc->buf_size - desc->buf_end;
            if (n < size)
            {
                _bufio_set_minbuf(desc, size + desc->buf_end /* - desc->buf_start */);
                n = desc->buf_size - desc->buf_end;
                ASSERT(n >= size);
            }
            dst = desc->buf + desc->buf_end;
            memmove(dst, buf, size);
            desc->buf_end += size;
            desc->flags |= BUFIO_F_WRBUF_DIRTY;
            n = size;
        }
        else
        {
            n = 0;
        }
        ASSERT(desc->buf_start <= desc->buf_end);
        ASSERT(desc->buf_start <= desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);
    }
    else if (size > 0)
    {
        ssize_t l;
        int r = 1;

        errno = 0;

        if (bufio_is_sock(desc))
        {
            DEBUG_DEVEL(DBGvars("going to wait for write event on sock = %d (timeout = %ld)\n"), desc->fd, cfg.write_timeout);
            r = tl_selectw(desc->fd, (int)cfg.write_timeout);
            DEBUG_DEVEL(DBGvars("waited for write event on sock = %d --> %d\n"), desc->fd, r);
        }
        if (r > 0)
        {
            l = _bufio_tl_write(desc, (char *)buf, size);
            if (errno != 0 || l <= 0)
            {
                ASSERT(errno != 0);
                if (!errno)
                    errno = EIO;
                desc->bufio_errno = errno;
            }
            n = l;
        }
    }
    desc->flags &= ~BUFIO_F_RD_WAS_LAST;
    desc->flags |= BUFIO_F_WR_WAS_LAST;

    ASSERT((desc->flags & BUFIO_F_RD_WAS_LAST) ? !(desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);
    ASSERT((desc->flags & BUFIO_F_WRBUF_DIRTY) ? (desc->flags & BUFIO_F_WR_WAS_LAST) : TRUE);

    return n;
}

static ssize_t _bufio_read(bufio *desc, char *buf, size_t size)
{
    ssize_t read_sz = 0;
    ssize_t read_act;
    size_t miss_size = size;
    ssize_t acnt;

    if (desc->eof)
    {
        return 0;
    }

    if (desc->bufio_errno)
    {
        errno = desc->bufio_errno;
        return -1;
    }

    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_RD_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return -1;
    }

    for (;;)
    {
        if (desc->buf_start < desc->buf_end)
        {
            read_act = TL_MIN((ssize_t)miss_size, desc->buf_end - desc->buf_start);

            memcpy(buf + read_sz, desc->buf + desc->buf_start, read_act);

            desc->buf_start += read_act;
            ASSERT(desc->buf_start <= desc->buf_end);
            ASSERT(desc->buf_start <= desc->buf_size);
            ASSERT(desc->buf_end <= desc->buf_size);
            miss_size -= read_act;
            read_sz += read_act;

            if (read_sz == size)
            {
                break;          /* return read_sz; */
            }
        }
        else
        {
            desc->buf_start = 0;
            desc->buf_end = 0;
            acnt = _bufio_tl_read(desc, desc->buf, desc->buf_size);
            if (acnt <= 0)
            {
                if (acnt == 0)
                {
                    desc->eof = 1;
                }
                else
                {
                    ASSERT(errno);
                    if (!errno)
                        errno = EIO;
                    desc->bufio_errno = errno;
                }
                if (read_sz)
                {
                    break;      /* return read_sz; */
                }
                else
                {
                    read_sz = acnt;
                    break;
                }
            }

            desc->buf_end = acnt;
            ASSERT(desc->buf_start <= desc->buf_end);
            ASSERT(desc->buf_start <= desc->buf_size);
            ASSERT(desc->buf_end <= desc->buf_size);
        }
    }

    desc->flags &= ~BUFIO_F_WR_WAS_LAST;
    desc->flags |= BUFIO_F_RD_WAS_LAST;

    return read_sz;
}

static ssize_t _bufio_nbfread(bufio *desc, char *buf, size_t size)
{
    ssize_t read_sz;

    if (desc->eof)
    {
        return 0;
    }
    if (desc->bufio_errno)
    {
        errno = desc->bufio_errno;
        return -1;
    }

    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_RD_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return -1;
    }

    if (desc->buf_start < desc->buf_end)
    {
        read_sz = TL_MIN((ssize_t)size, desc->buf_end - desc->buf_start);

        memcpy(buf, desc->buf + desc->buf_start, read_sz);

        desc->buf_start += read_sz;
        ASSERT(desc->buf_start <= desc->buf_end);
        ASSERT(desc->buf_start <= desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);

        /* return read_sz; */
    }
    else
    {
        desc->buf_start = 0;
        desc->buf_end = 0;
        read_sz = _bufio_tl_read(desc, buf, size);
        if (read_sz <= 0)
        {
            if (read_sz == 0)
            {
                desc->eof = 1;
            }
            if (read_sz < 0)
            {
                desc->bufio_errno = errno;
            }
        }
        /* return read_sz; */
    }

    desc->flags &= ~BUFIO_F_WR_WAS_LAST;
    desc->flags |= BUFIO_F_RD_WAS_LAST;

    return read_sz;
}

static ssize_t _bufio_readln(bufio *desc, char *buf, size_t size)
{
    ssize_t read_sz = 0;
    ssize_t read_act;
    size_t miss_size = size - 1;
    ssize_t acnt;
    char *mc = NULL;

    if (desc->eof || !size)
    {
        return 0;
    }

    if (desc->bufio_errno)
    {
        errno = desc->bufio_errno;
        return -1;
    }

    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_RD_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return -1;
    }

    size--;

    for (;;)
    {
        if (desc->buf_start < desc->buf_end)
        {
            mc = memchr(desc->buf + desc->buf_start, '\n', desc->buf_end - desc->buf_start);

            if (mc && (mc - (desc->buf + desc->buf_start) < miss_size))
            {
                read_act = mc + 1 - (desc->buf + desc->buf_start);
            }
            else
            {
                read_act = TL_MIN((ssize_t)miss_size, desc->buf_end - desc->buf_start);
            }
            memcpy(buf + read_sz, desc->buf + desc->buf_start, read_act);

            desc->buf_start += read_act;
            ASSERT(desc->buf_start <= desc->buf_end);
            ASSERT(desc->buf_start <= desc->buf_size);
            ASSERT(desc->buf_end <= desc->buf_size);
            miss_size -= read_act;
            read_sz += read_act;

            if (read_sz == size || mc)
            {
                *(buf + read_sz) = '\0';

                break;          /* return read_sz; */
            }
        }
        else
        {
            desc->buf_start = 0;
            desc->buf_end = 0;
            acnt = _bufio_tl_read(desc, desc->buf, desc->buf_size);

            if (acnt <= 0)
            {
                if (acnt == 0)
                {
                    desc->eof = 1;
                }
                else
                {
                    ASSERT(errno);
                    if (!errno)
                        errno = EIO;
                    desc->bufio_errno = errno;
                }
                if (read_sz)
                {
                    *(buf + read_sz) = '\0';

                    break;      /* return read_sz; */
                }
                else
                {
                    read_sz = acnt;
                    break;
                }
            }

            desc->buf_end = acnt;
            ASSERT(desc->buf_start <= desc->buf_end);
            ASSERT(desc->buf_start <= desc->buf_size);
            ASSERT(desc->buf_end <= desc->buf_size);
        }
    }

    desc->flags &= ~BUFIO_F_WR_WAS_LAST;
    desc->flags |= BUFIO_F_RD_WAS_LAST;

    return read_sz;
}

static void _bufio_unread(bufio *desc, char *buf, size_t size)
{
    if (_bufio_check_and_fix_last_action(desc, BUFIO_F_RD_WAS_LAST))
    {
        /* error condition! */
        ASSERT(desc->bufio_errno);
        errno = desc->bufio_errno;
        return;
    }

    VERIFY(!_bufio_fixup_buf(desc, size, 0));

    ASSERT(desc->buf_start >= size);
    if (desc->buf_start >= size && size > 0)
    {
        ASSERT(desc->buf);
        ASSERT(desc->buf_size >= size);
        ASSERT(desc->buf_start <= desc->buf_end);
        ASSERT(desc->buf_start <= desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);
        memmove(desc->buf + desc->buf_start - size, buf, size);
        desc->buf_start -= size;
        ASSERT(desc->buf_start <= desc->buf_end);
        ASSERT(desc->buf_start <= desc->buf_size);
        ASSERT(desc->buf_end <= desc->buf_size);
    }

    desc->flags &= ~BUFIO_F_WR_WAS_LAST;
    desc->flags |= BUFIO_F_RD_WAS_LAST;
}



ssize_t bufio_read(bufio *desc, char *buf, size_t size)
{
    ssize_t retv;

    _bufio_mutex_lock(desc);

    retv = _bufio_read(desc, buf, size);

    _bufio_mutex_unlock(desc);

    return retv;
}

ssize_t bufio_nbfread(bufio *desc, char *buf, size_t size)
{
    ssize_t retv;

    _bufio_mutex_lock(desc);

    retv = _bufio_nbfread(desc, buf, size);

    _bufio_mutex_unlock(desc);

    return retv;
}

ssize_t bufio_readln(bufio *desc, char *buf, size_t size)
{
    ssize_t retv;

    _bufio_mutex_lock(desc);

    retv = _bufio_readln(desc, buf, size);

    _bufio_mutex_unlock(desc);

    return retv;
}

void bufio_unread(bufio *desc, char *buf, size_t size)
{
    _bufio_mutex_lock(desc);

    _bufio_unread(desc, buf, size);

    _bufio_mutex_unlock(desc);
}




#if defined(ASSERT_IS_ENABLED)

/*
 * This one runs the risk of triggering ASSERT() statements within those bufio_*()
 * and xprintf() calls, so we use a crude check to make sure we don't coredump due
 * to infinite recursion.
 */
void tld_assert(const char *sourcefile, int lineno, const char *msg)
{
    static volatile int depth = 0;

#ifdef HAVE_MT
    int limit = cfg.nthr;
#else
    int limit = 1;
#endif
    if (depth <= limit)
    {
        depth++;
        if (!msg || !msg[0])
        {
            msg = "---";
        }
        printf("(%s:%d) ASSERTION (%s) FAILED!\n", sourcefile, lineno, msg);
        xprintf(1, "(%s:%d) ASSERTION (%s) FAILED!\n", sourcefile, lineno, msg);
        xprintf(1, gettext("Program has been forcefully terminated\n"));
        if (bufio_is_open(cfg.logfile))
        {
            _bufio_flush(cfg.logfile, bufio_is_open_file(cfg.logfile));
        }
        cfg.retcode_override = PAVUK_EXIT_ASSERT_ERR;
#if defined(I_FACE)
        cfg.xi_face = FALSE;
#endif
        cfg.rbreak = TRUE;
        /* cfg.stop = TRUE; */
        tl_sleep(1);
        exit(PAVUK_EXIT_ASSERT_ERR);
    }
    else
    {
        printf("(%s:%d) ASSERTION (%s) FAILED AT OVER-depth %d/%d!\n", sourcefile, lineno, msg, depth, limit);
        cfg.retcode_override = PAVUK_EXIT_ASSERT_ERR;
#if defined(I_FACE)
        cfg.xi_face = FALSE;
#endif
        cfg.rbreak = TRUE;
        /* cfg.stop = TRUE; */
        abort();
    }
}

#endif

