/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "css.h"
#include "url.h"
#include "tools.h"
#include "mode.h"

#define BUFFER_STEP_SIZE        1024
#define FENDER_SIZE             64

/*
 * Make sure (by resizing) that the given buffer '*buf' at size 'bufsize'
 * can store 'bump_up_amount' extra characters, while 'offset' characters have
 * already been stored in it.
 *
 * Return the (new) bufsize as a return value.
 */
static inline int expand_buffer(char **buf, int bufsize, int offset, int bump_up_amount)
{
    if (bufsize < (bump_up_amount + offset + FENDER_SIZE))
    {
        bufsize += bump_up_amount + BUFFER_STEP_SIZE;
        *buf = _realloc(*buf, bufsize);
    }
    return bufsize;
}



/*
 * Extract all URLs from the CSS file.
 *
 * NOTE: this code does not mind about CSS (*...*) ('('=='/' ;-) ) comments,
 * so you'll also get any URLs which have been commented out. The benefit
 * here is that any CSS comment-based Mac IE CSS hacks are processed as if we
 * were MSIE itself; the drawback is that the code does not listen to any
 * '-noread_comments' commandline option; that's why the documentation of
 * that one says explicitly it is meant for HTML comments.
 *
 * This filter routine does NOT modify the CSS data/file input 'stylestr'.
 */
dllist *css_get_all_links(url *doc_url, const char *stylestr, const char *base, const char *baset, int no_limits)
{
    bool_t urlstart = FALSE;
    bool_t importstart = FALSE;
    const char *p;
    char *stack;
    int sr, ssz;
    dllist *retval = NULL;
    char terminator = 0;

    ssz = 2 * BUFFER_STEP_SIZE;
    stack = _malloc(ssz);
    sr = 0;

    for (p = stylestr; *p; p++)
    {
        if (urlstart)
        {
            if (*p == terminator)
            {
                char *pom;
                char *sp = stack;

                if (*p == ')')
                {
                    char *ep;

                    /* p--;  -- no need to unwind as we are looking for URLs only */

                    stack[sr] = '\0';

                    ep = stack + sr - 1;
                    while (ep > stack && tl_ascii_isspace(*ep))
                    {
                        *ep = '\0';
                        ep--;
                    }
                    while (tl_ascii_isspace(*sp))
                        sp++;
                    /* start and end quote MUST match! */
                    if (ep > sp && *sp == *ep && (*sp == '\"' || *sp == '\''))
                    {
                        terminator = *sp;
                        *ep = '\0';
                        sp++;
                    }
                }
                else
                {
                    stack[sr] = '\0';

                    sp = stack;
                }

                pom = url_to_absolute_url(base, baset, doc_url, sp);
                if (pom)
                {
                    url *purl;
                    cond_info_t condp = { 0 };

                    /*
                     * level == 0 (CL0): condition checking is called from CSS parsing
                     * routine.
                     */
                    condp.level = 0;
                    condp.urlnr = 0;
                    condp.size = 0;
                    condp.time = 0L;
                    condp.mimet = NULL;
                    condp.full_tag = NULL;
                    condp.params = NULL;
                    condp.html_doc = stylestr;
                    condp.html_doc_offset = p - stylestr;
                    condp.tag = NULL;
                    condp.attrib = NULL;

                    purl = url_parse(pom);
                    ASSERT(purl->type != URLT_FROMPARENT);
                    LOCK_URL(purl);
                    purl->parent_url = dllist_append(purl->parent_url, (dllist_t)doc_url);
                    purl->level = doc_url->level + 1;
                    UNLOCK_URL(purl);
                    purl->status |= URL_INLINE_OBJ;
                    url_path_abs(purl);

                    if ((cfg.mode == MODE_SYNC || cfg.mode == MODE_MIRROR) && (purl->type == URLT_FILE))
                    {
                        url *pomurl = filename_to_url(purl->p.file.filename);

                        if (pomurl)
                        {
                            free_deep_url(purl);
                            _free(purl);
                            purl = pomurl;
                        }
                    }

                    if (no_limits || url_append_condition(purl, &condp))
                    {
                        retval = dllist_append(retval, (dllist_t)purl);
                    }
                    else
                    {
                        free_deep_url(purl);
                        _free(purl);
                    }
                    _free(pom);
                }

                urlstart = FALSE;
                importstart = FALSE;
                sr = 0;
            }
            else
            {
                ssz = expand_buffer(&stack, ssz, sr, 1);
                stack[sr] = *p;
                sr++;
            }
            continue;
        }

        /* @import MUST be followed by a [double]quote-delimited URL: */
        if (importstart && (*p == '\'' || *p == '"'))
        {
            urlstart = TRUE;
            terminator = *p;
            sr = 0;
            continue;
        }
        if (((p - stylestr) > 3) && !strncasecmp(p - 3, "url(", 4))
        {
            urlstart = TRUE;
            terminator = ')';
            sr = 0;
        }
        else if (!urlstart && !strncasecmp(p, "@import", 7))
        {
            importstart = TRUE;
            p += 7 - 1;
        }
    }

    _free(stack);
    return retval;
}


/*
 * Convert any URL in there to an 'absolute' URL and rewrite the CSS data.
 *
 * The 'rewritten' CSS data/file is passed back to the caller in the
 * function return value.
 *
 * NOTES:
 *
 * #1: what was said at 'css_get_all_links()' applies here too: this code
 *     does not consider CSS comments as such.
 *
 * #2: since the rewriting assumes a certain CSS standards adherence, you
 *     may find that some @import and CSS comment-based 'CSS hacks' will
 *     be destroyed or damaged. This code does NOT attempt to keep all
 *     known (and unknown) CSS hacks out there intact. Hence, when
 *     copying sites like that, YMMV. Or you'd have to specify '-noRelocate'
 *     on the command line.
 *
 *
 * extra: inspect usage of these functions: rewrite_links(),
 *        rewrite_one_parent_links(): both call this code with unconditionally
 *        turning on 'rewrite', so those are disregarding the
 *        cfg.rewrite_links setting!
 *
 *        See the rewrite-regardless code in html_process_parent_document()
 */
char *css_to_absolute_links(url *doc_url, const char *stylestr, const char *base, const char *baset)
{
    bool_t urlstart = FALSE;
    bool_t importstart = FALSE;
    char *result;
    int sz, r;
    const char *p;
    char *stack;
    int ssz, sr, l;
    char terminator = 0;

    sz = strlen(stylestr) + BUFFER_STEP_SIZE;
    result = _malloc(sz);
    r = 0;

    ssz = 2 * BUFFER_STEP_SIZE;
    stack = _malloc(ssz);
    sr = 0;

    for (p = stylestr; *p; p++)
    {
        if (urlstart)
        {
            if (*p == terminator)
            {
                char *pom;
                char *sp = stack;

                if (*p == ')')
                {
                    char *ep;

                    stack[sr] = '\0';

                    ep = stack + sr - 1;
                    while (ep > stack && tl_ascii_isspace(*ep))
                    {
                        *ep = '\0';
                        ep--;
                    }
                    while (tl_ascii_isspace(*sp))
                        sp++;
                    /* start and end quote MUST match! */
                    if (ep > sp && *sp == *ep && (*sp == '\"' || *sp == '\''))
                    {
                        terminator = *sp;
                        *ep = '\0';
                        sp++;
                    }
                }
                else
                {
                    stack[sr] = '\0';

                    sp = stack;
                }

                pom = url_to_absolute_url(base, baset, doc_url, sp);

                if (pom)
                {
                    url *purl;

                    purl = url_parse(pom);
                    ASSERT(purl->type != URLT_FROMPARENT);
                    _free(pom);
                    url_path_abs(purl);
                    pom = url_to_urlstr(purl, TRUE);
                    free_deep_url(purl);
                    _free(purl);
                }
                else
                {
                    pom = tl_strdup(sp);
                }

                l = strlen(pom);
                sz = expand_buffer(&result, sz, r, l + 3);

                /*
                 * some CSS style sheets use ' and/or " for dealing with filenames
                 * which contain (illegal?) quotes themselves. Then there's the
                 * url(<url>) without any quotes at all as well.
                 * Try to keep this intact.
                 */
                if (*p == ')')
                {
                    p--;        /* unwind ONE character so that the copy code further below will copy the ')' matching 'url(' */
                }

                /*
                 * when the 'terminator' has not not yet been assigned some quote,
                 * we are sure there were no quotes to begin with
                 */
                if (terminator != ')')
                {
                    /* when the URL contains one type of quote, try to cover for this by the other quote type */
                    if (strchr(pom, terminator))
                    {
                        if (terminator == '\'')
                            terminator = '"';
                        else
                            terminator = '\'';
                    }
                    result[r] = terminator;
                    r++;
                }
                strcpy(result + r, pom);
                r += l;
                if (terminator != ')')
                {
                    result[r] = terminator;
                    r++;
                }
                /* closing brace for 'url(...)' will be copied later, thanks to 'p--' above */

                urlstart = FALSE;
                importstart = FALSE;
                sr = 0;

                _free(pom);
            }
            else
            {
                ssz = expand_buffer(&stack, ssz, sr, 1);
                stack[sr] = *p;
                sr++;
            }
            continue;
        }

        /* @import MUST be followed by a [double]quote-delimited URL: */
        if (importstart && (*p == '\'' || *p == '"'))
        {
            urlstart = TRUE;
            terminator = *p;
            sr = 0;
            continue;
        }
        if (((p - stylestr) > 3) && !strncasecmp(p - 3, "url(", 4))
        {
            urlstart = TRUE;
            terminator = ')';
            sr = 0;
        }
        else if (!urlstart && !strncasecmp(p, "@import", 7))
        {
            importstart = TRUE;
        }
        sz = expand_buffer(&result, sz, r, 1);
        result[r] = *p;
        r++;
    }

    /*
     * Now what if the code still thinks it's collecting a URL, due to a badly formatted CSS file?
     * Cover that eventuality here:
     */
    if (sr)
    {
        sz = expand_buffer(&result, sz, r, sr);

        memcpy(result + r, stack, sr);
        r += sr;
    }
    result[r] = '\0';

    _free(stack);

    return result;
}


/*
 * Convert any URL in there to a 'local' version and rewrite the CSS data.
 *
 * The 'rewritten' CSS data/file is passed back to the caller in the
 * function return value.
 *
 * NOTES:
 *
 * what was said at 'css_to_absolute_links()' applies here too: this code
 * does not consider CSS comments as such, and rewriting may damage
 * particular so-called 'CSS hacks' (see quircksmode.com for elaboration
 * of that particular subject. 'ALA' a.k.a. 'A List Apart' is another site
 * containing quite a bit of info on this.
 *
 *
 * extra: inspect usage of these functions: rewrite_links(),
 *        rewrite_one_parent_links(): both call this code with unconditionally
 *        turning on 'rewrite', so those are disregarding the
 *        cfg.rewrite_links setting!
 *
 *        See the rewrite-regardless code in html_process_parent_document()
 */
char *css_remote_to_local_links(url *doc_url, const char *stylestr, int all, int sel, const char *base, const char *baset)
{
    bool_t urlstart = FALSE;
    bool_t importstart = FALSE;
    char *result;
    int sz, r;
    const char *p;
    const char *act_name = url_to_filename(doc_url, TRUE);
    const char *anchor;
    char *stack;
    int ssz, sr, l;
    char terminator = 0;

    sz = strlen(stylestr) + BUFFER_STEP_SIZE;
    result = _malloc(sz);
    r = 0;

    ssz = 2 * BUFFER_STEP_SIZE;
    stack = _malloc(ssz);
    sr = 0;

    for (p = stylestr; *p; p++)
    {
        if (urlstart)
        {
            if (*p == terminator)
            {
                char *pom;
                int pomsize;
                const char *fn;
                url *urlp;
                url *purl;
                struct stat estat = { 0 };
                char *sp = stack;

                if (*p == ')')
                {
                    char *ep;

                    stack[sr] = '\0';

                    ep = stack + sr - 1;
                    while (ep > stack && tl_ascii_isspace(*ep))
                    {
                        *ep = '\0';
                        ep--;
                    }
                    while (tl_ascii_isspace(*sp))
                        sp++;
                    /* start and end quote MUST match! */
                    if (ep > sp && *sp == *ep && (*sp == '\"' || *sp == '\''))
                    {
                        terminator = *sp;
                        *ep = '\0';
                        sp++;
                    }
                }
                else
                {
                    stack[sr] = '\0';

                    sp = stack;
                }

                urlp = url_parse(sp);
                ASSERT(urlp->type != URLT_FROMPARENT);
                if (all && urlp->type == URLT_FILE)
                {
                    fn = url_to_absolute_url(base, baset, doc_url, sp);
                    free_deep_url(urlp);
                    _free(urlp);
                    urlp = url_parse(fn);
                    ASSERT(urlp->type != URLT_FROMPARENT);
                }
                anchor = url_get_anchor_name(urlp);

                /** for better performance with info files **/
                if ((purl = url_was_befor(urlp)))
                    fn = url_to_filename(purl, TRUE);
                else
                    fn = url_to_filename(urlp, FALSE);

                pomsize = strlen(sp) + 1;
                pomsize = TL_MAX(pomsize, TL_MAX(4096, PATH_MAX));
                pom = _malloc(pomsize);
                strcpy(pom, sp);

                if (((all || (sel && purl)) && fn) || (!sel
                                                      && (((urlp->type != URLT_FILE) && fn && ((!tl_access(fn, R_OK) && !tl_stat(fn, &estat)) || !strcmp(act_name, fn))))))
                {
                    if (all || (sel && purl) || (!sel && ((!tl_access(fn, R_OK) && !S_ISDIR(estat.st_mode)) || !strcmp(act_name, fn))))
                    {
                        char *relfn = get_relative_path(act_name, fn);
                        if (anchor)
                            snprintf(pom, pomsize, "%s#%s", relfn, anchor);
                        else
                            strncpy(pom, relfn, pomsize);
                        pom[pomsize - 1] = 0;

                        _free(relfn);
                    }
                    else
                    {
                        snprintf(pom, pomsize, "%s%s%s", fn, (tl_is_dirname(fn) ? "" : "/"), priv_cfg.index_name);
                        pom[pomsize - 1] = 0;

                        if (!tl_stat(pom, &estat) && !S_ISDIR(estat.st_mode))
                        {
                            char *relfn = get_relative_path(act_name, pom);

                            if (anchor)
                                snprintf(pom, pomsize, "%s#%s", relfn, anchor);
                            else
                                strncpy(pom, relfn, pomsize);
                            pom[pomsize - 1] = 0;

                            _free(relfn);
                        }
                    }
                }
                free_deep_url(urlp);
                _free(urlp);

                l = strlen(pom);
                sz = expand_buffer(&result, sz, r, l + 3);

                /*
                 * some CSS style sheets use ' and/or " for dealing with filenames
                 * which contain (illegal?) quotes themselves. Then there's the
                 * url(<url>) without any quotes at all as well.
                 * Try to keep this intact.
                 */
                if (*p == ')')
                {
                    p--;        /* unwind ONE character so that the copy code further below will copy the ')' matching 'url(' */
                }

                /*
                 * when the 'terminator' has not not yet been assigned some quote,
                 * we are sure there were no quotes to begin with
                 */
                if (terminator != ')')
                {
                    /* when the URL contains one type of quote, try to cover for this by the other quote type */
                    if (strchr(pom, terminator))
                    {
                        if (terminator == '\'')
                            terminator = '"';
                        else
                            terminator = '\'';
                    }
                    result[r] = terminator;
                    r++;
                }
                strcpy(result + r, pom);
                r += l;
                if (terminator != ')')
                {
                    result[r] = terminator;
                    r++;
                }
                /* closing brace for 'url(...)' will be copied later, thanks to 'p--' above */

                urlstart = FALSE;
                importstart = FALSE;
                sr = 0;

                _free(pom);
            }
            else
            {
                ssz = expand_buffer(&stack, ssz, sr, 1);
                stack[sr] = *p;
                sr++;
            }
            continue;
        }

        /* @import MUST be followed by a [double]quote-delimited URL: */
        if (importstart && (*p == '\'' || *p == '"'))
        {
            urlstart = TRUE;
            terminator = *p;
            sr = 0;
            continue;
        }
        if (((p - stylestr) > 3) && !strncasecmp(p - 3, "url(", 4))
        {
            urlstart = TRUE;
            terminator = ')';
            sr = 0;
        }
        else if (!urlstart && !strncasecmp(p, "@import", 7))
        {
            importstart = TRUE;
        }
        sz = expand_buffer(&result, sz, r, 1);
        result[r] = *p;
        r++;
    }

    /*
     * Now what if the code still thinks it's collecting a URL, due to a badly formatted CSS file?
     * Cover that eventuality here:
     */
    if (sr)
    {
        sz = expand_buffer(&result, sz, r, sr);

        memcpy(result + r, stack, sr);
        r += sr;
    }
    result[r] = '\0';

    _free(stack);

    return result;
}




/*
 * Replace any use of the URL 'url_old' with the new URL 'url_new'.
 *
 * The 'rewritten' CSS data/file is passed back to the caller in the
 * function return value.
 *
 * NOTES:
 *
 * what was said at 'css_to_absolute_links()' applies here too: this code
 * does not consider CSS comments as such, and rewriting may damage
 * particular so-called 'CSS hacks' (see quircksmode.com for elaboration
 * of that particular subject. 'ALA' a.k.a. 'A List Apart' is another site
 * containing quite a bit of info on this.
 *
 *
 * extra: inspect usage of these functions: rewrite_links(),
 *        rewrite_one_parent_links(): both call this code with unconditionally
 *        turning on 'rewrite', so those are disregarding the
 *        cfg.rewrite_links setting!
 *
 *        See the rewrite-regardless code in html_process_parent_document()
 */
char *css_change_url(url *doc_url, const char *stylestr, url *url_old, const char *url_new)
{
    bool_t importstart = FALSE;
    bool_t urlstart = FALSE;
    char *result;
    int sz, r;
    const char *p;
    const char *anchor;
    char *oldfn;
    char *stack;
    int ssz, sr, l;
    char terminator = 0;

    sz = strlen(stylestr) + BUFFER_STEP_SIZE;
    result = _malloc(sz);
    r = 0;

    ssz = 2 * BUFFER_STEP_SIZE;
    stack = _malloc(ssz);
    sr = 0;

    oldfn = url_to_urlstr(url_old, FALSE);

    for (p = stylestr; *p; p++)
    {
        if (urlstart)
        {
            if (*p == terminator)
            {
                char *pom;
                int pomsize;
                char *pfn;
                url *urlp;
                char *sp = stack;

                if (*p == ')')
                {
                    char *ep;

                    stack[sr] = '\0';

                    ep = stack + sr - 1;
                    while (ep > stack && tl_ascii_isspace(*ep))
                    {
                        *ep = '\0';
                        ep--;
                    }
                    while (tl_ascii_isspace(*sp))
                        sp++;
                    /* start and end quote MUST match! */
                    if (ep > sp && *sp == *ep && (*sp == '\"' || *sp == '\''))
                    {
                        terminator = *sp;
                        *ep = '\0';
                        sp++;
                    }
                }
                else
                {
                    stack[sr] = '\0';

                    sp = stack;
                }

                pfn = NULL;
                urlp = url_parse(sp);
                ASSERT(urlp->type != URLT_FROMPARENT);
                anchor = url_get_anchor_name(urlp);

                pomsize = strlen(sp) + 1;
                pomsize = TL_MAX(pomsize, TL_MAX(4096, PATH_MAX));
                pom = _malloc(pomsize);
                strcpy(pom, sp);

                if (urlp && (urlp->type != URLT_FILE) && (pfn = url_to_urlstr(urlp, FALSE)))
                {
                    if (!strcmp(pfn, oldfn))
                    {
                        if (anchor)
                            snprintf(pom, pomsize, "%s#%s", url_new, anchor);
                        else
                            strncpy(pom, url_new, pomsize);
                        pom[pomsize - 1] = 0;
                    }
                }

                free_deep_url(urlp);
                _free(urlp);
                _free(pfn);

                l = strlen(pom);
                sz = expand_buffer(&result, sz, r, l + 3);

                /*
                 * some CSS style sheets use ' and/or " for dealing with filenames
                 * which contain (illegal?) quotes themselves. Then there's the
                 * url(<url>) without any quotes at all as well.
                 * Try to keep this intact.
                 */
                if (*p == ')')
                {
                    p--;        /* unwind ONE character so that the copy code further below will copy the ')' matching 'url(' */
                }

                /*
                 * when the 'terminator' has not not yet been assigned some quote,
                 * we are sure there were no quotes to begin with
                 */
                if (terminator != ')')
                {
                    /* when the URL contains one type of quote, try to cover for this by the other quote type */
                    if (strchr(pom, terminator))
                    {
                        if (terminator == '\'')
                            terminator = '"';
                        else
                            terminator = '\'';
                    }
                    result[r] = terminator;
                    r++;
                }
                strcpy(result + r, pom);
                r += l;
                if (terminator != ')')
                {
                    result[r] = terminator;
                    r++;
                }
                /* closing brace for 'url(...)' will be copied later, thanks to 'p--' above */

                urlstart = FALSE;
                importstart = FALSE;
                sr = 0;

                _free(pom);
            }
            else
            {
                ssz = expand_buffer(&stack, ssz, sr, 1);
                stack[sr] = *p;
                sr++;
            }
            continue;
        }

        /* @import MUST be followed by a [double]quote-delimited URL: */
        if (importstart && (*p == '\'' || *p == '"'))
        {
            urlstart = TRUE;
            terminator = *p;
            sr = 0;
            continue;
        }
        if (((p - stylestr) > 3) && !strncasecmp(p - 3, "url(", 4))
        {
            urlstart = TRUE;
            terminator = ')';
            sr = 0;
        }
        else if (!urlstart && !strncasecmp(p, "@import", 7))
        {
            importstart = TRUE;
        }
        sz = expand_buffer(&result, sz, r, 1);
        result[r] = *p;
        r++;
    }

    /*
     * Now what if the code still thinks it's collecting a URL, due to a badly formatted CSS file?
     * Cover that eventuality here:
     */
    if (sr)
    {
        sz = expand_buffer(&result, sz, r, sr);

        memcpy(result + r, stack, sr);
        r += sr;
    }
    result[r] = '\0';

    _free(stack);
    _free(oldfn);

    return result;
}

