/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "abstract.h"
#include "doc.h"
#include "errcode.h"
#include "gopher.h"
#include "http.h"
#include "net.h"
#include "url.h"
#include "tools.h"


/********************************************************/
/*
 * Send gopher request to the server.
 *
 * Return the socket bufio of the open connection, which will
 * receive the server reponse.
 *
 * Return NULL on error.
 */
/********************************************************/
bufio *gopher_get_data_socket(doc *docu)
{
    char pom[2048];
    const char *host;
    int port;

    DEBUG_PROCS(NULL);

    if (priv_cfg.gopher_proxy)
    {
        host = priv_cfg.gopher_proxy;
        port = cfg.gopher_proxy_port;
    }
    else
    {
        host = docu->doc_url->p.gopher.host;
        port = docu->doc_url->p.gopher.port;
    }

    docu->datasock = bufio_sock_fdopen(net_connect(host, port, docu), O_RDWR);
    if (docu->datasock)
    {
        if (priv_cfg.gopher_proxy)
        {
      if(http_dummy_proxy_connect(docu,
          url_get_site(docu->doc_url),
          url_get_port(docu->doc_url), host, port))
            {
                docu->errcode = ERR_PROXY_CONNECT;
                bufio_close(docu->datasock);
                docu->datasock = NULL;

                DEBUG_PROCE(NULL);
                return NULL;
            }
        }

        DEBUG_HTML("Gopher complete selector: '%s'\n", docu->doc_url->p.gopher.selector);

        snprintf(pom, sizeof(pom), "%s\r\n", docu->doc_url->p.gopher.selector + 1);
        pom[sizeof(pom) - 1] = 0;

#ifdef HAVE_DEBUG_FEATURES
        if (cfg.debug && (DEBUG_PROTOC_BITMASK & cfg.debug_level))
        {
            char dumpbuf[TEXTDUMP_BUFSIZE];

            DEBUG_PROTOC("********************* Gopher request (len = %ld) ********\n"
                         "%s\n"
                         "***************************************************\n",
                         (long)strlen(pom), tl_asciidump_content(dumpbuf, NUM_ELEM(dumpbuf), pom, strlen(pom)));
        }
#endif
        if (abs_write(docu->datasock, pom, strlen(pom)) != strlen(pom))
        {
            xperror(gettext("Gopher request transmit failure"));
            bufio_close(docu->datasock);

            DEBUG_PROCE(NULL);
            return NULL;
        }
    }
    else
    {
        docu->errcode = ERR_GOPHER_CONNECT;
        if (_h_errno_)
        {
            xherror(host);
        }
        else
        {
            xperror("net_connect");
        }
    }

    DEBUG_PROCE(NULL);

    return docu->datasock;
}


/********************************************************/
/*
 * Transform the received gopher data into a HTML page, URLs and all.
 */
/********************************************************/
void gopher_dir_to_html(doc *docu)
{
    char *p;
    char *res = NULL;
    char tp;
    const char *title;
    const char *host;
    char *sel;
    int port;
    int ilen;
    bool_t last = 1;

    DEBUG_PROCS(NULL);

    if (docu->doc_url->p.gopher.selector[0] == '1')
    {
        res = tl_str_printf(res,
                            gettext("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
                                    "<HTML>\n<TITLE>\nDirectory of gopher://%s:%hu/%s\n</TITLE>\n"
                                    "<BODY>\n<H1 ALIGN=CENTER><B>"
                                    "Directory of gopher://%s:%hu/%s </H1><PRE>\n"),
                            docu->doc_url->p.gopher.host, docu->doc_url->p.gopher.port,
    docu->doc_url->p.gopher.selector, docu->doc_url->p.gopher.host,
    docu->doc_url->p.gopher.port, docu->doc_url->p.gopher.selector);

        ASSERT(docu->contents ? docu->contents[docu->size] == 0 : TRUE);
        p = docu->contents;

        while (*p)
        {
            const char **fields;

            ilen = strcspn(p, "\r\n");
            if (p[ilen])
            {
                p[ilen] = '\0';
            }
            else
            {
                last = 0;
            }

            if (*p == '.')
            {
                break;
            }

            DEBUG_DEVEL(DBGvars("treating gopher data line '%s'...\n"), p);

            fields = (const char **)tl_str_split(p, "\t");

            if (tl_strv_length(fields) > 3)
            {
                char *augmented_title;

                title = fields[0];
                tp = *title;
                title++;

                sel = url_encode_str(fields[1], URL_PATH_UNSAFE);
                host = fields[2];
                port = tl_atoi(fields[3]);

                if (*title)
                {
                    augmented_title = tl_strdup(title);
                }
                else
                {
                    /* empty title: generate one... */
                    augmented_title = tl_str_printf(NULL, "(%s:%d%s%s)", host, port, (*sel != '/' ? "/" : ""), sel);
                }

                DEBUG_DEVEL(DBGvars("--> title: '%s' (augmented: '%s'), tp: %d (= '%c'), selector: '%s', host: '%s', port: %d\n"),
                            title, augmented_title, tp, tp, sel, host, port);

                /*
                 *      Gopher types (RFC 1436 + Wikipedia):
                 *
                 * File-types are described in gopher menus by a single letter or number.
                 * The original protocol defines 14 types, with others being added by
                 * the community.
                 * Older clients may not handle new file types, such as d for PDF, which is
                 * why many authors use the generic 9 for all binary files, hoping that
                 * the client's computer will be able to correctly process the file.
                 *
                 * The client should interpret the type field as follows:
                 *
                 * 0   The item is a TextFile Entity.
                 *     Client should use a TextFile Transaction.
                 *
                 * 1   The item is a Menu Entity.
                 *     Client should use a Menu Transaction.
                 *
                 * 2   The information applies to a CSO phone book entity.
                 *     Client should talk CSO protocol.
                 *
                 * 3   Signals an error condition.
                 *
                 * 4   Item is a Macintosh file encoded in BINHEX format
                 *
                 * 5   Item is PC-DOS binary file of some sort.  Client gets to decide.
                 *
                 * 6   Item is a uuencoded file.
                 *
                 * 7   The information applies to a Index Server.
                 *     Client should use a FullText Search transaction.
                 *
                 * 8   The information applies to a Telnet session.
                 *     Connect to given host at given port. The name to login as at this
                 *     host is in the selector string.
                 *
                 * 9   Item is a binary file.  Client must decide what to do with it.
                 *
                 * +   The information applies to a duplicated server.  The information
                 *     contained within is a duplicate of the primary server.  The primary
                 *     server is defined as the last DirEntity that is has a non-plus
                 *     "Type" field.  The client should use the transaction as defined by
                 *     the primary server Type field.
                 *
                 * g   Item is a GIF graphic file.
                 *
                 * I   Item is some kind of image file.  Client gets to decide.
                 *
                 * T   The information applies to a tn3270 based telnet session.
                 *     Connect to given host at given port. The name to login as at this
                 *     host is in the selector string.
                 *
                 * d   PDF file
                 *
                 * h   HTML file
                 *
                 * i   informational message
                 *
                 * s   audio file
                 *
                 * w - WWW address
                 *
                 * : - image   x
                 *
                 * ; - movie   x
                 *
                 * < - sound   x
                 */

                switch (tp)
                {
                case '2':
                    /*
                     * 2   The information applies to a CSO phone book entity.
                     *     Client should talk CSO protocol.
                     */
                    res = tl_str_printf(res, "<A HREF=\"cso://%s:%d/%c%s\">CSO: %s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case '3':
                    /*
                     * 3   Signals an error condition.
                     */
                    res = tl_str_printf(res, "<A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case '8':
                    /*
                     * 8   The information applies to a Telnet session.
                     *     Connect to given host at given port. The name to login as at this
                     *     host is in the selector string.
                     */
                    res = tl_str_printf(res,
                                        "<A HREF=\"telnet://%s@%s:%d/\">%s (user '%s', host: '%s', port: '%d')</A>\n",
                                        sel, host, port, augmented_title, sel, host, port);
                    break;

                case '+':
                    /*
                     * +   The information applies to a duplicated server.  The information
                     *     contained within is a duplicate of the primary server.  The primary
                     *     server is defined as the last DirEntity that has a non-plus
                     *     "Type" field.  The client should use the transaction as defined by
                     *     the primary server Type field.
                     */
                    res = tl_str_printf(res, "[SERVER] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case 'g':
                case 'I':
                case ':':
                    /*
                     * g   Item is a GIF graphic file.
                     */
                    /*
                     * I   Item is some kind of image file.  Client gets to decide.
                     *
                     * : - image   x
                     */
                    res = tl_str_printf(res,
                                        "<IMG SRC=\"gopher://%s:%d/%c%s\" ALT=\"%s\" TITLE=\"%s\">\n", host, port, tp, sel, augmented_title, augmented_title);
                    break;

                case 'T':
                    /*
                     * T   The information applies to a tn3270 based telnet session.
                     *     Connect to given host at given port. The name to login as at this
                     *     host is in the selector string.
                     */
                    res = tl_str_printf(res,
                                        "<A HREF=\"tn3270://%s@%s:%d/\">%s (user: '%s', host: '%s', port: %d)</A>\n",
                                        sel, host, port, augmented_title, sel, host, port);
                    break;

                case 'i':
                    /*
                     * i   informational message
                     */
                    res = tl_str_printf(res, "%s\n", title);
                    break;

                case '0':
                    /*
                     * 0   The item is a TextFile Entity.
                     *     Client should use a TextFile Transaction.
                     */
                    res = tl_str_printf(res, "[TEXT] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case '1':
                    /*
                     * 1   The item is a Menu Entity.
                     *     Client should use a Menu Transaction.
                     */
                    res = tl_str_printf(res, "[MENU] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case '4':
                    /*
                     * 4   Item is a Macintosh file encoded in BINHEX format
                     */
                    res = tl_str_printf(res, "[BINHEX] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case '5':
                    /*
                     * 5   Item is PC-DOS binary file of some sort.  Client gets to decide.
                     */
                    res = tl_str_printf(res, "[PC-DOS] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;


                case '6':
                    /*
                     * 6   Item is a uuencoded file.
                     */
                    res = tl_str_printf(res, "[UUENCODED] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case '7':
                    /*
                     * 7   The information applies to a Index Server.
                     *     Client should use a FullText Search transaction.
                     */
                    res = tl_str_printf(res,
                                        "[INDEX/FULL-TEXT SEARCH] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case '9':
                    /*
                     * 9   Item is a binary file.  Client must decide what to do with it.
                     */
                    res = tl_str_printf(res, "[BINARY DATA] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case 'd':
                    /*
                     * d   PDF file
                     */
                    res = tl_str_printf(res, "[PDF] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case 'h':
                case 'w':
                    /*
                     * h   HTML file: *selector* format is:
                     *
                     * 'URL:<uri>'
                     *
                     * w - WWW address
                     */
                    if (!strncmp(sel, "URL:", 4))
                    {
                        res = tl_str_printf(res, "[HTML] <A HREF=\"%s\">%s (URL: %s)</A>\n", sel + 4, title, sel + 4);
                    }
                    else
                    {
                        res = tl_str_printf(res, "<A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    }
                    break;

                case 's':
                case '<':
                    /*
                     * s   audio file
                     *
                     * < - sound   x
                     */
                    res = tl_str_printf(res, "[AUDIO] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                case ';':
                    /*
                     * ; - movie   x
                     */
                    res = tl_str_printf(res, "[VIDEO] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;

                default:
                    res = tl_str_printf(res, "[UNIDENTIFIED] <A HREF=\"gopher://%s:%d/%c%s\">%s</A>\n", host, port, tp, sel, augmented_title);
                    break;
                }

                _free(sel);
            }
            else
            {
                xprintf(1, gettext("Failed to parse Gopher directory entry:\n%s\n"), p);
            }

            p += ilen + last;
            p += strspn(p, "\r\n");

            tl_strv_free(fields);
        }

        res = tl_str_concat(DBGvars(res), "</PRE>\n</BODY>\n</HTML>\n", NULL);

        _free(docu->contents);

        docu->contents = res;
        docu->size = strlen(res);
    }

    DEBUG_PROCE(NULL);
}

