/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "html.h"
#include "tools.h"

/*************************************************************/
/*
 * This table defines all HTML tags recognized by pavuk, including any
 * supported (= processable) attribute.
 *
 * 'processable attributes' are those tag attributes, which may
 * contain URLs or JavaScript; basically anything which may be
 * important for a webpage-grabbing application like pavuk.
 *
 * Notes:
 *
 * <csobj> is a VERY-non-standard HTML tag, originating from Adobe GoLive.
 *         Pavuk supoorted it, and supports it still. I have no idea
 *         how complete our support for this tag is. (@ july 2008)
 *         However, improving that support beyond the basics is not a
 *         very high priority for me, to say the least.
 *
 *         See also
 *
 *           http://gadgetopia.com/post/3848
 *           http://code.google.com/webstats/2005-12/editors.html
 *
 *         Luckily, GoLive has been phased out:
 *
 *           http://www.adobe.com/products/golive/
 *
 *
 * <object> is often used to show Flash content. It comes with
 *         <param name="movie" value="my-url.swf" > tags in its body
 *         to point at the Flash movie.
 *
 *         Many sites use an extra non-standard tag <EMBED> with a
 *         'src="my-url.swf"' attribute to point at the same Flash
 *         movie, but not /all/ sites do this.
 *
 *         Nowadays, there's an growing third option (SWFobject
 *         JavaScript)): see here:
 *
 *           http://code.google.com/p/swfobject/wiki/documentation
 *
 * <embed> with a 'src="my-url"' attribute is non-standard (though
 *         HTML5 DRAFT supports it (see http://www.w3schools.com/tags/html5_embed.asp)
 *         and closely related to Flash / Audio / Video support
 *         on web sites in conjuction with the <OBJECT> tag; see
 *         above.
 *
 *
 * onload="...", etc.: script-enabled attributes.
 *
 *         See here for a complete list of these buggers:
 *
 *           http://www.w3.org/TR/html401/interact/scripts.html#events
 *
 *
 * Of general interest is the Google HTML survey from 2005:
 *
 * http://code.google.com/webstats/index.html
 *
 * Additionally, there's the HTML 3.2 and HTML 4.01 tag lists to
 * peruse:
 *
 * http://htmlhelp.com/reference/wilbur/alltags.html   (HTML 3.2)
 * http://www.w3.org/TR/REC-html40/index/elements.html (HTML 4.0)
 * http://www.wdvl.com/Authoring/HTML/4/Tags/Tags.html (HTML 4.01 - derived from W3C)
 *
 * And then there's HTML 5 (DRAFT!)
 *
 * http://dev.w3.org/html5/html-author/
 * http://www.whatwg.org/specs/web-apps/current-work/
 * http://www.w3schools.com/tags/html5.asp
 *
 */
/*
 * JavaScript events:
 *
 * onload = script [CT]
 *     This attribute may be used with BODY and FRAMESET elements.
 * onunload = script [CT]
 *     This attribute may be used with BODY and FRAMESET elements.
 * onclick = script [CT]
 *     This attribute may be used with most elements.
 * ondblclick = script [CT]
 *     This attribute may be used with most elements.
 * onmousedown = script [CT]
 *     This attribute may be used with most elements.
 * onmouseup = script [CT]
 *     This attribute may be used with most elements.
 * onmouseover = script [CT]
 *     This attribute may be used with most elements.
 * onmousemove = script [CT]
 *     This attribute may be used with most elements.
 * onmouseout = script [CT]
 *     This attribute may be used with most elements.
 * onfocus = script [CT]
 *     This attribute may be used with the following elements: A, AREA, LABEL, INPUT, SELECT, TEXTAREA, and BUTTON.
 * onblur = script [CT]
 *     This attribute may be used with the following elements: A, AREA, LABEL, INPUT, SELECT, TEXTAREA, and BUTTON.
 * onkeypress = script [CT]
 *     This attribute may be used with most elements.
 * onkeydown = script [CT]
 *     This attribute may be used with most elements.
 * onkeyup = script [CT]
 *     This attribute may be used with most elements.
 * onsubmit = script [CT]
 *     It only applies to the FORM element.
 * onreset = script [CT]
 *     It only applies to the FORM element.
 * onselect = script [CT]
 *     This attribute may be used with the INPUT and TEXTAREA elements.
 * onchange = script [CT]
 *     This attribute applies to the following elements: INPUT, SELECT, and TEXTAREA.
 */
#define COMMON_JAVASCRIPT_ATTRIBUTES                                                              \
    { HTML_ATTRIB_JSEVENT, "ONMOUSEOVER", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON }, \
    { HTML_ATTRIB_JSEVENT, "ONCLICK", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON },     \
    { HTML_ATTRIB_JSEVENT, "ONDBLCLICK", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON },  \
    { HTML_ATTRIB_JSEVENT, "ONMOUSEDOWN", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON }, \
    { HTML_ATTRIB_JSEVENT, "ONMOUSEUP", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON },   \
    { HTML_ATTRIB_JSEVENT, "ONMOUSEMOVE", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON }, \
    { HTML_ATTRIB_JSEVENT, "ONMOUSEOUT", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON },  \
    { HTML_ATTRIB_JSEVENT, "ONKEYPRESS", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON },  \
    { HTML_ATTRIB_JSEVENT, "ONKEYDOWN", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON },   \
    { HTML_ATTRIB_JSEVENT, "ONKEYUP", LINK_JS | LINK_DOWNLD | LINK_INLINE | LINK_JS_COMMON }

/*************************************************************/
html_tag_atrib_t html_common_attrib_tags[] =
{
	{HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
	COMMON_JAVASCRIPT_ATTRIBUTES,
};

html_tag_t html_link_tags[] = {
  {HTML_TAG_A, "A",
      {
          {HTML_ATTRIB_HREF, "HREF", LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONFOCUS", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONBLUR", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
		  {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_ABBR, "ABBR",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_ACRONYM, "ACRONYM",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_ADDRESS, "ADDRESS",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_APPLET, "APPLET",
      {
          {HTML_ATTRIB_CODEBASE, "CODEBASE", 0},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_AREA, "AREA",
      {
          {HTML_ATTRIB_HREF, "HREF", LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONFOCUS", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONBLUR", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_B, "B",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BASE, "BASE",
      {
          {HTML_ATTRIB_HREF, "HREF", 0},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BDO, "BDO",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BGSOUND, "BGSOUND",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BIG, "BIG",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BLOCKQUOTE, "BLOCKQUTE",
      {
          {HTML_ATTRIB_CITE, "CITE", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BODY, "BODY",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONLOAD", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONUNLOAD", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BR, "BR",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_BUTTON, "BUTTON",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONFOCUS", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONBLUR", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_CAPTION, "CAPTION",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_CENTER, "CENTER",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_CITE, "CITE",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_CODE, "CODE",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_COL, "COL",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_COLGROUP, "COLGROUP",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_CSOBJ, "CSOBJ",
      {
          {HTML_ATTRIB_HT, "HT", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_HT, "CSREF", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_DD, "DD",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_DEL, "DEL",
      {
          {HTML_ATTRIB_CITE, "CITE", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_DFN, "DFN",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_DIR, "DIR",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_DIV, "DIV",
      {
          {HTML_ATTRIB_HREF, "HREF", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_DL, "DL",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_DT, "DT",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_EM, "EM",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_EMBED, "EMBED",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_FIELDSET, "FIELDSET",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_FIG, "FIG",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_FONT, "FONT",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_FORM, "FORM",
      {
          {HTML_ATTRIB_ACTION, "ACTION", /* LINK_DOWNLD | */ LINK_FORM},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONSUBMIT", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONRESET", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_FRAME, "FRAME",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_LONGDESC, "LONGDESC", LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_FRAMESET, "FRAMESET",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONLOAD", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONUNLOAD", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_H1, "H1",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_H2, "H2",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_H3, "H3",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_H4, "H4",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_H5, "H5",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_H6, "H6",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_HEAD, "HEAD",
      {
          {HTML_ATTRIB_PROFILE, "PROFILE", LINK_INLINE | LINK_DOWNLD},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_HR, "HR",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_I, "I",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_IFRAME, "IFRAME",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_DOWNLD},
          {HTML_ATTRIB_LONGDESC, "LONGDESC", LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_IMG, "IMG",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_LOWSRC, "LOWSRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_LONGDESC, "LONGDESC", LINK_DOWNLD},
          {HTML_ATTRIB_USEMAP, "USEMAP", LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_INPUT, "INPUT",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_USEMAP, "USEMAP", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONSELECT", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONCHANGE", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONFOCUS", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONBLUR", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_INS, "INS",
      {
          {HTML_ATTRIB_CITE, "CITE", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_ISINDEX, "ISINDEX",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_KBD, "KBD",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_LABEL, "LABEL",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONFOCUS", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONBLUR", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_LAYER, "LAYER",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_LEGEND, "LEGEND",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_LI, "LI",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_LINK, "LINK",
      {
          {HTML_ATTRIB_HREF, "HREF", LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_MAP, "MAP",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_MENU, "MENU",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_META, "META",
      {
          {HTML_ATTRIB_CONTENT, "CONTENT", LINK_DOWNLD},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_NOFRAMES, "NOFRAMES",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_NOSCRIPT, "NOSCRIPT",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_OBJECT, "OBJECT",
      {
          {HTML_ATTRIB_DATA, "DATA", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_USEMAP, "USEMAP", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_OL, "OL",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_OPTGROUP, "OPTGROUP",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_OPTION, "OPTION",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_OVERLAY, "OVERLAY",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_P, "P",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  /* <param name="movie" value="myContent.swf" /> */
  {HTML_TAG_PARAM, "PARAM",
      {
          {HTML_ATTRIB_VALUE, "VALUE", LINK_INLINE | LINK_DOWNLD | LINK_SWF_OBJ_PARAM },
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_PRE, "PRE",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_Q, "Q",
      {
          {HTML_ATTRIB_CITE, "CITE", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_S, "S",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SAMP, "SAMP",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SCRIPT, "SCRIPT",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_SCRIPT | LINK_DOWNLD},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SELECT, "SELECT",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONCHANGE", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONFOCUS", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONBLUR", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SMALL, "SMALL",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SOUND, "SOUND",
      {
          {HTML_ATTRIB_SRC, "SRC", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SPAN, "SPAN",
      {
          {HTML_ATTRIB_HREF, "HREF", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_STRIKE, "STRIKE",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_STRONG, "STRONG",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SUB, "SUB",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_SUP, "SUP",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_TABLE, "TABLE",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_TBODY, "TBODY",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_TD, "TD",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_TEXTAREA, "TEXTAREA",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
          {HTML_ATTRIB_JSEVENT, "ONSELECT", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONCHANGE", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONFOCUS", LINK_JS | LINK_DOWNLD | LINK_INLINE},
          {HTML_ATTRIB_JSEVENT, "ONBLUR", LINK_JS | LINK_DOWNLD | LINK_INLINE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_TFOOT, "TFOOT",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_TH, "TH",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_THEAD, "THEAD",
      {
          {HTML_ATTRIB_BACKGROUND, "BACKGROUND", LINK_INLINE | LINK_DOWNLD},
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_TT, "TT",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_U, "U",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_UL, "UL",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
  {HTML_TAG_VAR, "VAR",
      {
          {HTML_ATTRIB_STYLE, "STYLE", LINK_STYLE},
		  COMMON_JAVASCRIPT_ATTRIBUTES,
          {HTML_ATTRIB_NULL, NULL, 0}
        }
    }
  ,
};


int html_link_tags_num(void)
{
    return NUM_ELEM(html_link_tags);
}

int html_common_attrib_tags_num(void)
{
    return NUM_ELEM(html_common_attrib_tags);
}

