/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "dns.h"
#include "gui_api.h"
#include "absio.h"
#include "net.h"
#include "doc.h"
#include "http.h"
#include "tools.h"


#ifdef HAVE_WINSOCK2_H

int WinSockErr2errno(int err)
{
    switch (err)
    {
    case 0:
        return 0;

    default:
        return EFAULT;

    case WSAEWOULDBLOCK:
        return EWOULDBLOCK;

    case WSAEINPROGRESS:
        return EINPROGRESS;

    case WSAEALREADY:
        return EALREADY;

    case WSAENOTSOCK:
        return ENOTSOCK;

    case WSAEDESTADDRREQ:
        return EDESTADDRREQ;

    case WSAEMSGSIZE:
        return EMSGSIZE;

    case WSAEPROTOTYPE:
        return EPROTOTYPE;

    case WSAENOPROTOOPT:
        return ENOPROTOOPT;

    case WSAEPROTONOSUPPORT:
        return EPROTONOSUPPORT;

    case WSAESOCKTNOSUPPORT:
        return ESOCKTNOSUPPORT;

    case WSAEOPNOTSUPP:
        return EOPNOTSUPP;

    case WSAEPFNOSUPPORT:
        return EPFNOSUPPORT;

    case WSAEAFNOSUPPORT:
        return EAFNOSUPPORT;

    case WSAEADDRINUSE:
        return EADDRINUSE;

    case WSAEADDRNOTAVAIL:
        return EADDRNOTAVAIL;

    case WSAENETDOWN:
        return ENETDOWN;

    case WSAENETUNREACH:
        return ENETUNREACH;

    case WSAENETRESET:
        return ENETRESET;

    case WSAECONNABORTED:
        return ECONNABORTED;

    case WSAECONNRESET:
        return ECONNRESET;

    case WSAENOBUFS:
        return ENOBUFS;

    case WSAEISCONN:
        return EISCONN;

    case WSAENOTCONN:
        return ENOTCONN;

    case WSAESHUTDOWN:
        return ESHUTDOWN;

    case WSAETOOMANYREFS:
        return ETOOMANYREFS;

    case WSAETIMEDOUT:
        return ETIMEDOUT;

    case WSAECONNREFUSED:
        return ECONNREFUSED;

    case WSAELOOP:
        return ELOOP;

    case WSAENAMETOOLONG:
        return ENAMETOOLONG;

    case WSAEHOSTDOWN:
        return EHOSTDOWN;

    case WSAEHOSTUNREACH:
        return EHOSTUNREACH;

    case WSAENOTEMPTY:
        return ENOTEMPTY;

    case WSAEPROCLIM:
        return EPROCLIM;

    case WSAEUSERS:
        return EUSERS;

    case WSAEDQUOT:
        return EDQUOT;

    case WSAESTALE:
        return ESTALE;

    case WSAEREMOTE:
        return EREMOTE;
    }
}


int net_host_to_in_addr(const char *hostname, abs_addr *haddr)
{
    int len;

    if (dns_gethostbyname(hostname, &len, haddr->addr, &haddr->family))
        return -1;

    return 0;
}

/*****************************************************************/
/*
 * Open a connection to the given hostname:port address and return
 * the socket handle on success or -1 on failure.
 *
 * Notes:
 *  'hostname' may be an IP address or domain name (which will be
 *             resolved to an IP address)
 *
 *  The socket for the opened (client) connection is opened in
 *  NONBLOCKING mode.
 *
 *  'docp' is passed to provide certain configuration values (proxy
 *  use, DNS request timeout value, URL type for hostname:port (HTTPS,
 *  HTTP, etc.).
 *
 *  Note also that 'docp' is used to STORE a trace
 *  record for the connection when running in 'hammer' mode. This
 *  trace record is used lateron to replay this connection while
 *  running in 'hammer' DoS mode.
 */
/*****************************************************************/
static int net_connect_helper(const char *hostname, int port_no, doc *docp)
{
    struct sockaddr *addr;
    struct sockaddr_storage saddr;
    abs_addr haddr;
    SOCKET sock;
    int rv, l, sas, tcp;

    int actual_port = port_no;
    const char *actual_hostname = hostname;
    http_proxy *hp;

    if (docp->doc_url->type == URLT_HTTPS && cfg.transparent_ssl_proxy)
    {
        hp = cfg.transparent_ssl_proxy;
    }
    else
    {
        hp = cfg.transparent_proxy;
    }

    /*
     * Check for transparent redirect
     */
    if (hp)
    {
        actual_hostname = hp->addr;
        actual_port = hp->port;
    }

    set_h_errno(0);

    if (net_host_to_in_addr(actual_hostname, &haddr))
        return -2;

    gettimeofday(&docp->dns_time, NULL);

    gui_loop_serve();

    tcp = dns_getprotoid("tcp");

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    if (docp->actrec_ref)
    {
        activity_record *rec = docp->actrec_ref;

        DEBUG_DEVEL(DBGvars("set host connect info #1: %d\n"), docp->doc_url->type);
        rec->type = docp->doc_url->type;
        rec->host_name = tl_strdup(actual_hostname);
        rec->host_port = actual_port;
        rec->socket_family = haddr.family;
        rec->socket_proto = tcp;
    }
#endif

    if ((sock = socket(haddr.family, SOCK_STREAM, tcp)) == INVALID_SOCKET)
    {
        errno = WinSockErr2errno(WSAGetLastError());
        xperror("socket()");
        return -1;
    }

    if (set_sock_as_nonblocking(sock, TRUE, TRUE))
    {
        /* the socket has already been closed by now. */
        xperror(gettext("set socket as nonblocking"));
        return -1;
    }

    if (cfg.local_ip)
    {
        struct sockaddr_storage sladdr;
        struct sockaddr *laddr;

        laddr = dns_setup_sockaddr(&cfg.local_ip_addr, 0, &sladdr, &sas);

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
        if (docp->actrec_ref)
        {
            activity_record *rec = docp->actrec_ref;

            DEBUG_DEVEL(DBGvars("set host connect local bind info\n"));
            rec->local_addr = _malloc(sas);
            memcpy(rec->local_addr, laddr, sas);
            rec->local_addr_len = sas;
        }
#endif

        if (bind(sock, laddr, sas) == SOCKET_ERROR)
        {
            errno = WinSockErr2errno(WSAGetLastError());
            xperror("bind() - local");
            closesocket(sock);
            return -1;
        }
    }

    addr = dns_setup_sockaddr(&haddr, actual_port, &saddr, &sas);

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    if (docp->actrec_ref)
    {
        activity_record *rec = docp->actrec_ref;

        DEBUG_DEVEL(DBGvars("set host connect dest addr info\n"));
        rec->dest_addr = _malloc(sas);
        memcpy(rec->dest_addr, addr, sas);
        rec->dest_addr_len = sas;
    }
#endif

    rv = connect(sock, addr, sas);
    errno = WinSockErr2errno(WSAGetLastError());
    if (rv == SOCKET_ERROR && (errno != EINPROGRESS) && (errno != EISCONN) && (errno != EAGAIN) && (errno != EWOULDBLOCK))
    {
        closesocket(sock);
        return -1;
    }


#if defined(I_FACE) && !defined(HAVE_MT)
    if (cfg.xi_face && cfg.done)
    {
        if (rv == SOCKET_ERROR && (errno == EINPROGRESS || errno == EAGAIN || errno == EWOULDBLOCK))
        {
            rv = gui_wait_io(sock, FALSE, cfg.conn_timeout);
            if (!rv)
            {
                DEBUG_NET("Async connect - connected\n");
            }

            if (cfg.rbreak || cfg.stop || rv)
            {
                closesocket(sock);
                return -1;
            }
        }
    }
    else
#endif /* !HAVE_MT */
    {
        if (rv == SOCKET_ERROR && (errno == EINPROGRESS || errno == EAGAIN || errno == EWOULDBLOCK))
        {
            DEBUG_DEVEL(DBGvars("going to wait for connect/write event on sock = %d (timeout = %ld)\n"), (int)sock, cfg.conn_timeout);
            while ((rv = tl_selectw(sock, cfg.conn_timeout)) == -1 && errno == EINTR && !cfg.stop && !cfg.rbreak)
            {
                DEBUG_DEVEL(DBGvars("waiting for connect/write event on sock = %d\n"), (int)sock);
            }
            if (rv <= 0)
            {
                if (rv == 0)
                    errno = ETIMEDOUT;
                closesocket(sock);
                return -1;
            }
        }
    }

    l = sizeof(rv);
    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (void *)&rv, &l) || rv)
    {
        closesocket(sock);
        errno = WinSockErr2errno(rv);
        return -1;
    }

    DEBUG_NET("Successfully connected to %s,%d\n", hostname, port_no);

    return (int)sock;
}

int net_connect(const char *hostname, int port_no, doc *docp)
{
    int rc = net_connect_helper(hostname, port_no, docp);

    /* if net_connect_helper() returned -2, this was dns
     * failure. If -1, it was connect() failure
     */
    if (rc == -2)
    {
        DEBUG_NET("DNS failure for %s,%d\n", hostname, port_no);
        rc = -1;
    }
    else
    {
        gettimeofday(&docp->connect_time, NULL);
    }
    return rc;
}

int net_bindport(abs_addr *haddr, int port_min, int port_max)
{
    SOCKET sock;
    int port;
    int n, tcp;
    struct sockaddr *addr;
    struct sockaddr_storage saddr;

    addr = (struct sockaddr *)&saddr;

    tcp = dns_getprotoid("tcp");

    if ((sock = socket(haddr->family, SOCK_STREAM, tcp)) == -1)
    {
        errno = WinSockErr2errno(WSAGetLastError());
        xperror("socket");
        return -1;
    }

    n = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&n, sizeof(n)))
    {
        errno = WinSockErr2errno(WSAGetLastError());
        xperror("net_bind: setsockopt - SO_REUSEADDR");
    }

    if (port_min < 0 || port_max < 0)
    {
        port = 0;

        addr = dns_setup_sockaddr(haddr, port, &saddr, &n);

        if (bind(sock, addr, n) != 0)
        {
            errno = WinSockErr2errno(WSAGetLastError());
            xvperror(DBGvars(NULL), "bind to port(%d)\n", port);
            closesocket(sock);
            return -1;
        }
    }
    else
    {
        /*
         * pro: take a port out of the specified range
         */
        int count = 50;
        double port_range = port_max - port_min + 1;
        int success = 0;

        DEBUG_NET("trying ports in range %d - %d\n", port_min, port_max);
        while (count >= 0)
        {
            if (count == 0)
            {
                port = 0;
                DEBUG_NET("trying system assigned port\n");
            }
            else
            {
                port = port_min + (int)(port_range * rand() / (RAND_MAX + 1.0));
                ASSERT(port >= port_min && port <= port_max);
            }

            addr = dns_setup_sockaddr(haddr, port, &saddr, &n);

            if (bind(sock, addr, n) == 0)
            {
                success = 1;
                break;
            }

            DEBUG_NET("trying new port : port %d failed\n", port);

            count--;
        }

        if (!success)
        {
            errno = WinSockErr2errno(WSAGetLastError());
            closesocket(sock);
            xvperror(DBGvars(NULL), "bind to ports (%d-%d)\n", port_min, port_max);
            return -1;
        }
    }

    n = sizeof(*addr);
    if (getsockname(sock, addr, &n))
    {
        errno = WinSockErr2errno(WSAGetLastError());
        closesocket(sock);
        xperror("getsockname");
        return -1;
    }

    DEBUG_NET("BINDING to port : %d\n", dns_get_sockaddr_port(addr));

    if (listen(sock, 1) == -1)
    {
        errno = WinSockErr2errno(WSAGetLastError());
        xperror("listen");
        closesocket(sock);
        return -1;
    }

    return (int)sock;
}

int net_accept(int sock)
{
    struct sockaddr *caller;
    struct sockaddr_storage scaller;
    int p;
    SOCKET rsock = INVALID_SOCKET;
    int rv;

    ASSERT(rsock < 0);
    if (set_sock_as_nonblocking(sock, TRUE, TRUE))
    {
        /* the socket has already been closed by now. */
        xperror(gettext("set socket as nonblocking"));
        return -1;
    }

    caller = (struct sockaddr *)&scaller;
    p = sizeof(scaller);

    rsock = accept(sock, caller, &p);
    errno = WinSockErr2errno(WSAGetLastError());
    if ((rsock < 0) && (errno != EWOULDBLOCK) && (errno != EAGAIN))
    {
        closesocket(sock);
        return -1;
    }

#if defined(I_FACE) && !defined(HAVE_MT)
    if (cfg.xi_face && cfg.done)
    {
        if ((rsock < 0) && (errno == EWOULDBLOCK || errno == EAGAIN))
        {
            rv = gui_wait_io(sock, TRUE, cfg.conn_timeout);
            if (rv || cfg.rbreak || cfg.stop)
            {
                closesocket(sock);
                return -1;
            }
        }
    }
    else
#endif /* !HAVE_MT */
    {
        if ((rsock < 0) && (errno == EWOULDBLOCK || errno == EAGAIN))
        {
            DEBUG_DEVEL(DBGvars("going to wait for connect event on sock = %d (timeout = %ld)\n"), sock, cfg.conn_timeout);
            while ((rv = tl_selectr(sock, cfg.conn_timeout)) == -1 && errno == EINTR && !cfg.stop && !cfg.rbreak)
            {
                DEBUG_DEVEL(DBGvars("waiting for connect event on sock = %d\n"), sock);
            }
            if (rv <= 0)
            {
                if (rv == 0)
                    errno = ETIMEDOUT;
                closesocket(sock);
                return -1;
            }
        }
    }

    if ((rsock < 0) && ((rsock = accept(sock, caller, &p)) == -1))
    {
        errno = WinSockErr2errno(WSAGetLastError());
        closesocket(sock);
        return -1;
    }

#ifdef HAVE_DEBUG_FEATURES
    if (cfg.debug && (DEBUG_NET_BITMASK & cfg.debug_level))
    {
        char *ip = dns_get_sockaddr_ip(caller);
        DEBUG_NET("ACCEPTING connection from: %s:%d\n", ip, dns_get_sockaddr_port(caller));
        _free(ip);
    }
#endif

    return (int)rsock;
}



/* [i_a] apr/2007 */
int sock_maybe_in_fdset(fd_set *set, int sock)
{
    int i;

    if (sock < 0)
        return FALSE;

    /* if set isn't filled up yet, this socket might fit in there. */
    if (set->fd_count < FD_SETSIZE)
        return TRUE;

    /* otherwise it MUST be in there */
    for (i = 0; i < set->fd_count; i++)
    {
        if (set->fd_array[i] == sock)
            return TRUE;
    }

    return FALSE;
}



/*
 * Return 0 on success.
 */
int set_sock_as_nonblocking(int sock, bool_t on, bool_t close_socket_on_error)
{
    u_long val;

    val = !!on;                 /* make sure val==1 or 0 */
    if (SOCKET_ERROR == ioctlsocket(sock, FIONBIO, &val))
    {
        errno = WinSockErr2errno(WSAGetLastError());
        if (close_socket_on_error)
        {
            closesocket(sock);
        }
        return -1;
    }
    return 0;
}




#endif /* HAVE_WINSOCK2_H */

