/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "dns.h"
#include "gui_api.h"
#include "absio.h"
#include "net.h"
#include "doc.h"
#include "http.h"
#include "tools.h"



#if !defined(HAVE_WINSOCK2_H)



int net_host_to_in_addr(const char *hostname, abs_addr *haddr)
{
    int len;

    if (dns_gethostbyname(hostname, &len, haddr->addr, &haddr->family))
        return -1;

    return 0;
}

/*****************************************************************/
/*
 * Open a connection to the given hostname:port address and return
 * the socket handle on success or -1 on failure.
 *
 * Notes:
 *  'hostname' may be an IP address or domain name (which will be
 *             resolved to an IP address)
 *
 *  The socket for the opened (client) connection is opened in
 *  NONBLOCKING mode.
 *
 *  'docp' is passed to provide certain configuration values (proxy
 *  use, DNS request timeout value, URL type for hostname:port (HTTPS,
 *  HTTP, etc.).
 *
 *  Note also that 'docp' is used to STORE a trace
 *  record for the connection when running in 'hammer' mode. This
 *  trace record is used lateron to reply this connection while
 *  running in 'hammer' DoS mode.
 */
/*****************************************************************/
static int net_connect_helper(const char *hostname, int port_no, doc *docp)
{
    struct sockaddr *addr;
    struct sockaddr_storage saddr;
    abs_addr haddr;
    int sock;
    int rv, sas, tcp;
    socklen_t socklen;

    int actual_port = port_no;
    const char *actual_hostname = hostname;
    http_proxy *hp;

    DEBUG_PROCS(NULL);

    if (docp->doc_url->type == URLT_HTTPS && cfg.transparent_ssl_proxy)
    {
        hp = cfg.transparent_ssl_proxy;
    }
    else
    {
        hp = cfg.transparent_proxy;
    }

    /*
     * Check for transparent redirect
     */
    if (hp)
    {
        actual_hostname = hp->addr;
        actual_port = hp->port;
    }

    set_h_errno(0);

    if (net_host_to_in_addr(actual_hostname, &haddr))
    {
        sock = -2;
        goto exit_stage_left;
    }

    gettimeofday(&docp->dns_time, NULL);

    gui_loop_serve();

    tcp = dns_getprotoid("tcp");

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    if (docp->actrec_ref)
    {
        activity_record *rec = docp->actrec_ref;

        DEBUG_DEVEL(DBGvars("set host connect info #1: %d\n"), docp->doc_url->type);
        rec->type = docp->doc_url->type;
        rec->host_name = tl_strdup(actual_hostname);
        rec->host_port = actual_port;
        rec->socket_family = haddr.family;
        rec->socket_proto = tcp;
    }
#endif

    if ((sock = socket(haddr.family, SOCK_STREAM, tcp)) == -1)
    {
        xperror("socket()");
        sock = -1;
        goto exit_stage_left;
    }

    if (set_sock_as_nonblocking(sock, TRUE, TRUE))
    {
        /* the socket has already been closed by now. */
        xperror(gettext("set socket as nonblocking"));
        sock = -1;
        goto exit_stage_left;
    }

    if (cfg.local_ip)
    {
        struct sockaddr_storage sladdr;
        struct sockaddr *laddr;

        laddr = dns_setup_sockaddr(&cfg.local_ip_addr, 0, &sladdr, &sas);

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
        if (docp->actrec_ref)
        {
            activity_record *rec = docp->actrec_ref;

            DEBUG_DEVEL(DBGvars("set host connect local bind info\n"));
            rec->local_addr = _malloc(sas);
            memcpy(rec->local_addr, laddr, sas);
            rec->local_addr_len = sas;
        }
#endif

        if (bind(sock, laddr, sas) != 0)
        {
            xperror("bind() - local");
            close(sock);
            sock = -1;
            goto exit_stage_left;
        }
    }

    addr = dns_setup_sockaddr(&haddr, actual_port, &saddr, &sas);

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    if (docp->actrec_ref)
    {
        activity_record *rec = docp->actrec_ref;

        DEBUG_DEVEL(DBGvars("set host connect dest addr info\n"));
        rec->dest_addr = _malloc(sas);
        memcpy(rec->dest_addr, addr, sas);
        rec->dest_addr_len = sas;
    }
#endif

    rv = connect(sock, addr, sas);
    if (rv && (errno != EINPROGRESS) && (errno != EISCONN) && (errno != EAGAIN) && (errno != EWOULDBLOCK))
    {
        close(sock);
        sock = -1;
        goto exit_stage_left;
    }


#if defined(I_FACE) && !defined(HAVE_MT)
    if (cfg.xi_face && cfg.done)
    {
        if (rv && (errno == EINPROGRESS || errno == EAGAIN || errno == EWOULDBLOCK))
        {
            rv = gui_wait_io(sock, FALSE, cfg.conn_timeout);
            if (!rv)
            {
                DEBUG_NET("Async connect - connected\n");
            }

            if (cfg.rbreak || cfg.stop || rv)
            {
                close(sock);
                sock = -1;
                goto exit_stage_left;
            }
        }
    }
    else
#endif /* !HAVE_MT */
    {
        if (rv && (errno == EINPROGRESS || errno == EAGAIN || errno == EWOULDBLOCK))
        {
            DEBUG_DEVEL(DBGvars("going to wait for connect/write event on sock = %d (timeout = %ld)\n"), sock, cfg.conn_timeout);
            while ((rv = tl_selectw(sock, cfg.conn_timeout)) == -1 && errno == EINTR && !cfg.stop && !cfg.rbreak)
            {
                DEBUG_DEVEL(DBGvars("waiting for connect/write event on sock = %d\n"), sock);
            }
            if (rv <= 0)
            {
                if (rv == 0)
                {
                    errno = ETIMEDOUT;
                }
                close(sock);
                sock = -1;
                goto exit_stage_left;
            }
        }
    }

#if defined __QNX__ || defined __BEOS__
    if (connect(sock, addr, sas) && (errno != EISCONN))
    {
        close(sock);
        sock = -1;
        goto exit_stage_left;
    }
#else
    socklen = sizeof(rv);
    if (getsockopt(sock, SOL_SOCKET, SO_ERROR, (void *)&rv, &socklen) || rv)
    {
        close(sock);
        errno = rv;
        sock = -1;
        goto exit_stage_left;
    }
#endif

    DEBUG_NET("Successfully connected to %s,%d\n", hostname, port_no);

exit_stage_left:
    DEBUG_PROCE(NULL);

    return sock;
}

int net_connect(const char *hostname, int port_no, doc *docp)
{
    int rc;

    DEBUG_PROCS(NULL);

    rc = net_connect_helper(hostname, port_no, docp);
    /* if net_connect_helper() returned -2, this was dns
     * failure. If -1, it was connect() failure
     */
    if (rc == -2)
    {
        DEBUG_NET("DNS failure for %s,%d\n", hostname, port_no);
        rc = -1;
    }
    else
    {
        gettimeofday(&docp->connect_time, NULL);
    }

    DEBUG_PROCE(NULL);

    return rc;
}

int net_bindport(abs_addr *haddr, int port_min, int port_max)
{
    int sock;
    int port;
    int n, tcp;
    struct sockaddr *addr;
    struct sockaddr_storage saddr;
    socklen_t socklen;

    DEBUG_PROCS(NULL);

    addr = (struct sockaddr *)&saddr;

    tcp = dns_getprotoid("tcp");

    if ((sock = socket(haddr->family, SOCK_STREAM, tcp)) == -1)
    {
        xperror("socket");
        sock = -1;
        goto exit_stage_left;
    }

    n = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&n, sizeof(n)))
    {
        xperror("net_bind: setsockopt - SO_REUSEADDR");
    }

    if (port_min < 0 || port_max < 0)
    {
        port = 0;

        addr = dns_setup_sockaddr(haddr, port, &saddr, &n);

        if (bind(sock, addr, n) != 0)
        {
            xvperror(DBGvars(NULL), "bind to port(%d)\n", port);
            close(sock);
            sock = -1;
            goto exit_stage_left;
        }
    }
    else
    {
        /*
         * pro: take a port out of the specified range
         */
        int count = 50;
        double port_range = port_max - port_min + 1;
        int success = 0;

        DEBUG_NET("trying ports in range %d - %d\n", port_min, port_max);
        while (count >= 0)
        {
            if (count == 0)
            {
                port = 0;
                DEBUG_NET("trying system assigned port\n");
            }
            else
            {
                port = port_min + (int)(port_range * rand() / (RAND_MAX + 1.0));
                ASSERT(port >= port_min && port <= port_max);
            }

            addr = dns_setup_sockaddr(haddr, port, &saddr, &n);

            if (bind(sock, addr, n) == 0)
            {
                success = 1;
                break;
            }

            DEBUG_NET("trying new port : port %d failed\n", port);

            count--;
        }

        if (!success)
        {
            close(sock);
            xvperror(DBGvars(NULL), "bind to ports (%d-%d)\n", port_min, port_max);
            sock = -1;
            goto exit_stage_left;
        }
    }

    socklen = sizeof(*addr);
    if (getsockname(sock, addr, &socklen))
    {
        close(sock);
        xperror("getsockname");
        sock = -1;
        goto exit_stage_left;
    }

    DEBUG_NET("BINDING to port : %d\n", dns_get_sockaddr_port(addr));

    if (listen(sock, 1) == -1)
    {
        xperror("listen");
        close(sock);
        sock = -1;
        goto exit_stage_left;
    }

exit_stage_left:
    DEBUG_PROCE(NULL);

    return sock;
}

int net_accept(int asock)
{
    struct sockaddr *caller;
    struct sockaddr_storage scaller;
    socklen_t p;
    int sock = -1;
    int rv;

    DEBUG_PROCS(NULL);

    if (set_sock_as_nonblocking(asock, TRUE, TRUE))
    {
        /* the socket has already been closed by now. */
        xperror(gettext("set accept socket as nonblocking"));
        sock = -1;
        goto exit_stage_left;
    }

    caller = (struct sockaddr *)&scaller;
    p = sizeof(scaller);

    sock = accept(asock, caller, &p);
    if ((sock < 0) && (errno != EWOULDBLOCK) && (errno != EAGAIN))
    {
        close(asock);
        sock = -1;
        goto exit_stage_left;
    }

#if defined(I_FACE) && !defined(HAVE_MT)
    if (cfg.xi_face && cfg.done)
    {
        if ((sock < 0) && (errno == EWOULDBLOCK || errno == EAGAIN))
        {
            rv = gui_wait_io(asock, TRUE, cfg.conn_timeout);
            if (rv || cfg.rbreak || cfg.stop)
            {
                close(asock);
                sock = -1;
                goto exit_stage_left;
            }
        }
    }
    else
#endif /* !HAVE_MT */
    {
        if ((sock < 0) && (errno == EWOULDBLOCK || errno == EAGAIN))
        {
            errno = 0;
            DEBUG_DEVEL(DBGvars("going to wait for connect event on sock = %d (timeout = %ld)\n"), asock, cfg.conn_timeout);
            while ((rv = tl_selectr(asock, cfg.conn_timeout)) == -1 && errno == EINTR && !cfg.stop && !cfg.rbreak)
            {
                DEBUG_DEVEL(DBGvars("waiting for connect event on sock = %d\n"), asock);
            }
            if (rv <= 0)
            {
                if (rv == 0)
                {
                    errno = ETIMEDOUT;
                }
                close(asock);
                sock = -1;
                goto exit_stage_left;
            }
        }
    }

    if ((sock < 0) && ((sock = accept(asock, caller, &p)) == -1))
    {
        close(asock);
        sock = -1;
        goto exit_stage_left;
    }

#ifdef HAVE_DEBUG_FEATURES
    if (cfg.debug && (DEBUG_NET_BITMASK & cfg.debug_level))
    {
        char *ip = dns_get_sockaddr_ip(caller);
        DEBUG_NET("ACCEPTING connection from: %s:%d\n", ip, dns_get_sockaddr_port(caller));
        _free(ip);
    }
#endif

exit_stage_left:
    DEBUG_PROCE(NULL);

    return sock;
}



/* [i_a] apr/2007 */
int sock_maybe_in_fdset(fd_set *set, int sock)
{
    return sock >= 0 && sock < FD_SETSIZE;
}



/*
 * Return 0 on success.
 */
int set_sock_as_nonblocking(int sock, bool_t on, bool_t close_socket_on_error)
{
    /*
     * This little number is derived from the OpenSSL code and personal proprietary code.
     * This mix should cover quite a few systems, including BSD and some items on or
     * over the edge regarding UNIXness. (VXWORKS, VMS, ...)
     *
     * For some interesting puzzles see the OpenSSL code expecting to use FIONBIO
     * for general nonblockiness (BIO_socket_nbio()), while, for instance, Mozilla's
     * nsprpub code employs O_NONBLOCK for the same in the UNIX zone (pt_MakeFdNonblock()).
     *
     * The top entry here is the most common one on UNIXes IMHO.
     */
#if defined(HAVE_FCNTL) && defined(F_GETFL) && defined(F_SETFL) && defined(O_NONBLOCK)
    int flags = fcntl(sock, F_GETFL, 0);
    if (flags == -1)
    {
        if (close_socket_on_error)
        {
            close(sock);
        }
        return -1;
    }
    flags &= ~O_NONBLOCK;
    if (on)
    {
        flags |= O_NONBLOCK;
    }
    if (fcntl(sock, F_SETFL, flags))
#elif defined(FIONBIO)
    int val;

    val = !!on;                 /* make sure val==1 or 0 */
#if defined(HAVE_IOCTLSOCKET)
    if (ioctlsocket(sock, FIONBIO, &val))
#elif defined(HAVE_IOCTL)
    if (ioctl(sock, FIONBIO, &val))
#elif defined(HAVE_SETSOCKOPT) && defined(SOL_SOCKET)
    if (setsockopt(sock, SOL_SOCKET, FIONBIO, &val, sizeof(val)))
#else
#error "Your system does not provide any known system calls to configure a socket handle as NONBLOCKING"
#endif
#endif
    {
        if (close_socket_on_error)
        {
            close(sock);
        }
        return -1;
    }
    return 0;
}




#endif /* !defined(HAVE_WINSOCK2_H) */

