/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#include "config.h"

#include "remind.h"

#include "bufio.h"
#include "url.h"
#include "mime.h"
#include "times.h"
#include "http.h"
#include "dllist.h"
#include "tools.h"
#include "gcinfo.h"
#include "gui_api.h"
#include "errcode.h"


static dllist *remind_urls = NULL;

static int remind_check_if_newer(doc *docp, time_t modtime, time_t *nmdtm)
{
    char *p;
    int retv = 0;
    http_response *rsp;
    char *ustr;

    DEBUG_PROCS(NULL);

    ustr = url_to_urlstr(docp->doc_url, FALSE);

    if (!cfg.quiet)
    {
        xprintf(1, gettext("Checking: %s\n"), ustr);
    }

    if (!(docp->doc_url->type == URLT_HTTP
#ifdef USE_SSL
         || docp->doc_url->type == URLT_HTTPS
#endif
         ))
    {
        xperror(gettext("This URL type is not supported in reminder mode\n"));
        retv = -1;
        goto exit_stage_left;
    }

    /*** to init properly proxy ... ***/
    doc_download_init(docp, FALSE);

    docp->request_type = HTTP_REQ_HEAD;
    docp->datasock = http_get_data_socket(docp);

    if (!docp->datasock)
    {
        retv = -1;
        goto exit_stage_left;
    }

    if (!docp->mime)
    {
        xperror(gettext("[remind] Bad response on HEAD request\n"));
        retv = -1;
        goto exit_stage_left;
    }

//printf("(%s:%d) txt = '%s'\n", __FILE__, __LINE__, docp->mime);
    rsp = http_get_response_info(docp->mime);

    if (cfg.use_http11 /* && !docu->http_proxy_10 */
       && rsp->ver_maj == 1 && rsp->ver_min == 1)
    {
        docp->is_http11 = TRUE;
        docp->is_persistent = TRUE;
    }
    else
    {
        docp->is_http11 = FALSE;
        docp->is_persistent = FALSE;
    }

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    if (docp->actrec_ref)
    {
        activity_record *rec = docp->actrec_ref;

        DEBUG_DEVEL(DBGvars("recording response code: %d\n"), rsp->ret_code);
        rec->expected_response_code = rsp->ret_code;
    }
#endif

    p = get_mime_param_val_str("Connection:", docp->mime);
    if (p)
    {
        if (!strcasecmp(p, "close"))
        {
            docp->is_persistent = FALSE;
        }
        else if (!strcasecmp(p, "keep-alive"))
        {
            docp->is_persistent = TRUE;
        }
        _free(p);
    }

#ifdef INCLUDE_CHUNKY_DoS_FEATURES
    if (docp->actrec_ref)
    {
        activity_record *rec = docp->actrec_ref;

        DEBUG_DEVEL(DBGvars("set persistent flag: %d\n"), docp->is_persistent);
        rec->expect_persistent = docp->is_persistent;
    }
#endif

    if (!docp->is_persistent)
    {
        bufio_close(docp->datasock);
        docp->datasock = NULL;
    }

    if (rsp->ret_code == HTTP_RSP_ACTUAL /* RFC2616: Section 10.3.5: Not Modified */)
    {
        retv = 0;
    }
    else if ((p = get_mime_param_val_str("Last-Modified:", docp->mime)))
    {
        bool_t sanitized = FALSE;

        *nmdtm = scntime(p, sanitized);
        if (*nmdtm == (time_t)-1)
        {
            xvperror(DBGvars(NULL), gettext("[remind] cannot decode Last-Modified date '%s' - will try to sanitize the timestamp.\n"), p);
            sanitized = TRUE;
            *nmdtm = scntime(p, sanitized);
        }
        if (*nmdtm == (time_t)-1)
        {
            xvperror(DBGvars(NULL), gettext("[remind] cannot decode Last-Modified date '%s'\n"), p);
            retv = -1;
        }
        else if (modtime && (difftime(*nmdtm, modtime) > 0))
        {
            retv = 1;
        }
#if defined(HAVE_DEBUG_FEATURES)
        if (cfg.debug && (DEBUG_TRACE_BITMASK & cfg.debug_level))
        {
            LOCK_TIME;
            DEBUG_TRACE("remind: Last-Modified date decoded: '%s'%s (original timestamp text: '%s')\n",
                        ctime(nmdtm), (sanitized ? " [sanitized!]" : ""), p);
            UNLOCK_TIME;
        }
#endif
        _free(p);
    }
    else
    {
        retv = -1;
    }

    _free(rsp->text);
    _free(rsp);

exit_stage_left:
    DEBUG_PROCE(NULL);

    return retv;
}

static void remind_check_entry(doc *docp, remind_entry *e)
{
    time_t t = 0L;
    int rv;

    rv = remind_check_if_newer(docp, e->mdtm, &t);

    if (!(cfg.stop || cfg.rbreak))
    {
        if (t)
        {
            e->mdtm = t;
        }

        if (rv == 1)
        {
            e->status |= REMIND_MODIFIED;
        }
        else if (rv == -1)
        {
            e->status |= REMIND_ERROR;
        }
    }
}

void remind_load_db(void)
{
    bufio *fd;
    char fnbuf[PATH_MAX + 1];
    char buf[BUFIO_ADVISED_READLN_BUFSIZE];

  snprintf(fnbuf, sizeof(fnbuf), "%s%s%s", cfg.path_to_home, 
	  (tl_is_dirname(cfg.path_to_home) ? "" : "/"), PAVUK_REMIND_DB_FILENAME);
    fnbuf[sizeof(fnbuf) - 1] = 0;

    while (remind_urls)
    {
        free_deep_url(((remind_entry *)remind_urls->data)->urlp);
        _free(remind_urls->data);
        remind_urls = dllist_remove_entry(remind_urls, remind_urls);
    }

    fd = bufio_copen(fnbuf, O_BINARY | O_RDONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    if (!fd)
    {
        xperror(mk_native(fnbuf));
        return;
    }

    bufio_flock(fd, TRUE);

    while (bufio_readln(fd, buf, sizeof(buf)) > 0)
    {
        time_t t;
        url *purl;
        char *eptr;
        remind_entry *entry;

        strip_nl(buf);
        t = strtol(buf, &eptr, 10);
        if (!t && errno == ERANGE)
        {
            xvperror(DBGvars(fnbuf), gettext("Bad reminder db entry - %s\n"), buf);
            continue;
        }

        while (*eptr && tl_ascii_isspace(*eptr))
            eptr++;
        purl = url_parse(eptr);

        entry = _malloc(sizeof(*entry));
        entry->urlp = purl;
        entry->mdtm = t;
        entry->status = 0;
        remind_urls = dllist_append(remind_urls, (dllist_t)entry);
    }

    bufio_funlock(fd);
    bufio_close(fd);
    fd = NULL;
}

void remind_save_db(void)
{
    bufio *fd;
    char buf[PATH_MAX + 1];
    dllist *ptr;

  snprintf(buf, sizeof(buf), "%s%s%s", cfg.path_to_home, 
	  (tl_is_dirname(cfg.path_to_home) ? "" : "/"), PAVUK_REMIND_DB_FILENAME);
    buf[sizeof(buf) - 1] = 0;

    fd = bufio_copen(buf, O_BINARY | O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    if (!fd)
    {
        xperror(mk_native(buf));
        return;
    }

    bufio_flock(fd, TRUE);

    bufio_truncate(fd, 0);
    for (ptr = remind_urls; ptr; ptr = ptr->next)
    {
        char *p;

        p = url_to_urlstr(((remind_entry *)ptr->data)->urlp, FALSE);
        bufio_printf(fd, "%ld %s\n", ((remind_entry *)ptr->data)->mdtm, p);
        _free(p);
    }

    bufio_funlock(fd);
    bufio_close(fd);
    fd = NULL;
}

static remind_entry *remind_find_entry(url *urlp)
{
    dllist *ptr;
    remind_entry *rv = NULL;
    char *ustr;

    ptr = remind_urls;

    ustr = url_to_urlstr(urlp, FALSE);

    while (ptr && !rv)
    {
        remind_entry *e = (remind_entry *)ptr->data;
        char *p;

        p = url_to_urlstr(e->urlp, FALSE);

        if (!strcmp(ustr, p))
        {
            rv = e;
        }
        _free(p);
        ptr = ptr->next;
    }
    _free(ustr);

    return rv;
}

void remind_start_add(void)
{
    dllist *ptr = cfg.request;

    while (ptr)
    {
        url_info *ui = (url_info *)ptr->data;
        url *urlp = url_parse(ui->urlstr);

        if (!remind_find_entry(urlp))
        {
            remind_entry *e = _malloc(sizeof(*e));

            e->urlp = urlp;
            e->status = 0;
            e->mdtm = 0L;

            remind_urls = dllist_append(remind_urls, (dllist_t)e);
        }
        else
        {
            free_deep_url(urlp);
            _free(urlp);
        }

        ptr = ptr->next;
    }
}

#ifdef HAVE_MT
static void _remind_do(int thnr)
#else
void remind_do(void)
#endif
{
    dllist *ptr;
    global_connection_info con_info;

    init_global_connection_data(&con_info);

    for (ptr = remind_urls; ptr; ptr = ptr->next)
    {
        doc docu;
        remind_entry *e;

        e = (remind_entry *)ptr->data;

        if (e->status & REMIND_PROCESSED)
            continue;

        doc_init(&docu, e->urlp);
#ifdef HAVE_MT
        docu.threadnr = thnr;
        VERIFY(0 == pthread_setspecific(cfg.currdoc_key, &docu));
        VERIFY(0 == pthread_setspecific(cfg.herrno_key, &docu.__herrno));
#endif
#ifdef INCLUDE_CHUNKY_DoS_FEATURES
        docu.cyclenr = 0;
#endif
        restore_global_connection_data(&con_info, &docu);

        remind_check_entry(&docu, e);

        save_global_connection_data(&con_info, &docu);

#ifdef HAVE_MT
        doc_finish_processing(&docu);
#endif

#ifdef HAVE_MT
        /*
         * as 'docu' is stack-based storage, make sure we destroy references pointing to that
         * storage once we unroll the stack that much. And that's right HERE!
         */
        VERIFY(0 == pthread_setspecific(cfg.currdoc_key, NULL));
        VERIFY(0 == pthread_setspecific(cfg.herrno_key, NULL));
#endif

        if (cfg.stop || cfg.rbreak)
        {
            break;
        }
        else
        {
            e->status |= REMIND_PROCESSED;
        }
    }

    kill_global_connection_data(&con_info);

    DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "remind_do() restored cfg.stop/cfg.rbreak", cfg.stop, cfg.rbreak);
}

#ifdef HAVE_MT
static void *mt_remind_do_thrd(void *param)
{
    int thrnr = cast_voidptr2int(param);

#ifdef I_FACE
    _config_struct_priv_t privcfg;

#if defined(__OSF__) || defined(__osf__)
#define __builtin_try
#define __builtin_finally
#endif
#endif

    {
#if defined(HAVE_SIGEMPTYSET) && defined(HAVE_SIGADDSET) && defined(HAVE_PTHREAD_SIGMASK)
        sigset_t smask;

        sigemptyset(&smask);
        sigaddset(&smask, SIGINT);
#ifdef SIGQUIT
        sigaddset(&smask, SIGQUIT);
#endif
#ifdef SIGXFSZ
        sigaddset(&smask, SIGXFSZ);
#endif
        pthread_sigmask(SIG_UNBLOCK, &smask, NULL);
#endif

        signal(SIGINT, pavuk_sigintthr);
#ifdef SIGQUIT
        signal(SIGQUIT, pavuk_sigquitthr);
#endif
#ifdef SIGXFSZ
        signal(SIGXFSZ, pavuk_sigfilesizethr);
#endif
    }

    VERIFY(0 == pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL));
#if defined(_MSC_VER)
    /*
     * warning C4312: 'type cast' : conversion from 'int' to 'void *' of greater size
     * warning C4312: 'type cast' : conversion from 'long' to 'void *' of greater size
     */
#pragma warning(push)
#pragma warning(disable: 4311 4312)
#endif
    VERIFY(0 == pthread_setspecific(cfg.currdoc_key, NULL));
    VERIFY(0 == pthread_setspecific(cfg.thrnr_key, cast_int2voidptr(thrnr)));
#if defined(_MSC_VER)
#pragma warning(pop)
#endif
    DEBUG_MTTHR(DBGvars("starting thread(%ld) %d\n"), pthread_t2long(pthread_self()), thrnr);

#ifdef I_FACE
    privcfg_make_copy(&privcfg);
    VERIFY(0 == pthread_setspecific(cfg.privcfg_key, &privcfg));
    pthread_cleanup_push(&privcfg_free4cleanup, &privcfg);
#endif

    for (; !cfg.rbreak && !cfg.stop;)
    {
        VERIFY(0 == pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL));
        DEBUG_MTTHR(DBGvars("thread %d awaking\n"), thrnr);

        _remind_do(thrnr);
        DEBUG_MTTHR(DBGvars("thread %d sleeping\n"), thrnr);
        gui_clear_status();
        gui_set_status(gettext("Sleeping ..."));
        VERIFY(0 == pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL));
        DEBUG_MTTHR(DBGvars("mt_semaphore_up(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
        mt_semaphore_up(&cfg.nrunning_sem);
        /*
         * UN-critical section; also signal that this thread isn't really active now.
         *
         * The effect of this is that the main thread can monitor if anybody is still
         * working on some URL; if all threads are waiting in the next loop for another
         * URL to arrive, the main thread will discover this after a while and will then
         * be able to signal all threads to stop as everything is done.
         */

        DEBUG_MTTHR(DBGvars("mt_semaphore_timed_wait(cfg.urlstack_sem: v = %ld BEFORE)\n"), cfg.urlstack_sem.v);
        while (!cfg.stop && !cfg.rbreak && mt_semaphore_timed_wait(&cfg.urlstack_sem, SUBSECOND_SEMA_POOL_TIMEOUT) < 0)
        {
      DEBUG_MTTHR(DBGvars("remind: [%d/%ld] wait for cfg.urlstack_sem: v = %ld\n"),
                  thrnr,
                  cfg.nthr,
                  cfg.urlstack_sem.v);
        }

        DEBUG_MTTHR(DBGvars("mt_semaphore_decrement(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
        mt_semaphore_decrement(&cfg.nrunning_sem);
    }
    DEBUG_MTTHR(DBGvars("mt_semaphore_up(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
    mt_semaphore_up(&cfg.nrunning_sem);

#ifdef I_FACE
    pthread_cleanup_pop(TRUE);
#endif
    DEBUG_MTTHR(DBGvars("thread %d exiting\n"), thrnr);
    gui_set_status(gettext("Exiting ..."));
    pthread_exit(NULL);
    return NULL;
}

void remind_do(void)
{
    pthread_attr_t thrdattr;
    int i;
    int num = cfg.nthr;

    DEBUG_PROCS(NULL);

    {
#if defined(HAVE_SIGEMPTYSET) && defined(HAVE_SIGADDSET) && defined(HAVE_PTHREAD_SIGMASK)
        sigset_t smask;

        sigemptyset(&smask);
        sigaddset(&smask, SIGINT);
#ifdef SIGQUIT
        sigaddset(&smask, SIGQUIT);
#endif
#ifdef SIGXFSZ
        sigaddset(&smask, SIGXFSZ);
#endif
        pthread_sigmask(SIG_UNBLOCK, &smask, NULL);
#endif

        signal(SIGINT, pavuk_sigintthr);
#ifdef SIGQUIT
        signal(SIGQUIT, pavuk_sigquitthr);
#endif
#ifdef SIGXFSZ
        signal(SIGXFSZ, pavuk_sigfilesizethr);
#endif
    }

    VERIFY(0 == pthread_attr_init(&thrdattr));
    VERIFY(0 == pthread_attr_setscope(&thrdattr, PTHREAD_SCOPE_SYSTEM));
    VERIFY(0 == pthread_attr_setstacksize(&thrdattr, MT_STACK_SIZE));

    DEBUG_MTTHR(DBGvars("mt_semaphore_init(nrunning/urlstack)\n"));
    mt_semaphore_destroy(&cfg.nrunning_sem);
    mt_semaphore_destroy(&cfg.urlstack_sem);
    DEBUG_MTTHR(DBGvars("mt_semaphore_init(cfg.nrunning_sem = 1)\n"));
    mt_semaphore_initv(&cfg.nrunning_sem, 1);
    DEBUG_MTTHR(DBGvars("mt_semaphore_init(cfg.urlstack_sem = 0)\n"));
    mt_semaphore_initv(&cfg.urlstack_sem, 0);
    /* signal each URL available in the queue */
    {
        dllist *dptr;

        LOCK_CFG_URLSTACK;
        DEBUG_MTTHR(DBGvars("multiple mt_semaphore_up(cfg.urlstack_sem: v = %ld BEFORE)\n"), cfg.urlstack_sem.v);
        for (dptr = cfg.urlstack; dptr; dptr = dptr->next)
        {
            mt_semaphore_up(&cfg.urlstack_sem);
        }
        DEBUG_MTTHR(DBGvars("multiple mt_semaphore_up(cfg.urlstack_sem: v = %ld AFTERWARDS)\n"), cfg.urlstack_sem.v);
        UNLOCK_CFG_URLSTACK;
    }

    if (num <= 0)
        num = 1;

    cfg.allthreadsnr = 0;
    cfg.allthreads = _malloc(num * sizeof(cfg.allthreads[0]));

    for (i = 0; i < num; i++)
    {
        if (!pthread_create(&cfg.allthreads[cfg.allthreadsnr], &thrdattr, &mt_remind_do_thrd, cast_int2voidptr(cfg.allthreadsnr)))
        {
            cfg.allthreadsnr++;
#if defined I_FACE
            gui_mt_thread_start(cfg.allthreadsnr);
#endif
            DEBUG_MTTHR(DBGvars("mt_semaphore_decrement(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
            mt_semaphore_decrement(&cfg.nrunning_sem);
        }
        else
        {
            xvperror(DBGvars(NULL), gettext("Create downloading thread %d\n"), i);
        }

        if (cfg.rbreak || cfg.stop)
            break;
    }

    DEBUG_MTTHR(DBGvars("mt_semaphore_timed_wait(cfg.nrunning_sem: v = %ld BEFORE)\n"), cfg.nrunning_sem.v);
    while (!cfg.stop && !cfg.rbreak && mt_semaphore_timed_wait(&cfg.nrunning_sem, SUBSECOND_SEMA_POOL_TIMEOUT) < 0)
    {
      DEBUG_MTTHR(DBGvars("remind: [-/%d] wait for cfg.nrunning_sem: v = %ld\n"),
                  num,
                  cfg.nrunning_sem.v);
    }

    DEBUG_DEVEL(DBGvars("%s cfg.stop = %d, cfg.rbreak = %d\n"), "remind (after waiting for cfg.running_sem) - ", cfg.stop, cfg.rbreak);
    /*
     * signal all processing threads to stop as each one of them is only waiting
     * for another URL to arrive in the URL stack/queue right now - which will never
     * happen as all URLs have been processed.
     */
    cfg.stop = TRUE;

    tl_msleep(SUBSECOND_SEMA_POOL_TIMEOUT, TRUE);

    for (i = 0; i < cfg.allthreadsnr; i++)
    {
#if 0
        pthread_cancel(cfg.allthreads[i]);
        pthread_kill(cfg.allthreads[i], SIGQUIT);
#endif
        pthread_join(cfg.allthreads[i], NULL);
    }

#if 0
    DEBUG_MTTHR(DBGvars("mt_semaphore_destroy(cfg.nrunning_sem)\n"));
    mt_semaphore_destroy(&cfg.nrunning_sem);
#endif

    _free(cfg.allthreads);
    cfg.allthreadsnr = 0;

#if defined I_FACE
    gui_mt_thread_end(0);
#endif

    DEBUG_PROCE(NULL);
}
#endif

void remind_send_result(void)
{
    FILE *fp;
    char buf[PATH_MAX];
    const char *p;
    char *dp;

    DEBUG_PROCS(NULL);

    if (cfg.remind_cmd)
    {
        for (dp = buf, p = cfg.remind_cmd; *p; p++)
        {
            if (*p == '%')
            {
                p++;
                switch (*(p))
                {
                case '%':
                    *dp = *p;
                    dp++;
                    break;

                case 'e':
                    strcpy(dp, priv_cfg.from);
                    while (*dp)
                    {
                        dp++;
                    }
                    break;

                case 'd':
                    {
                        time_t t = time(NULL);
                        char pom[100];

                        LOCK_TIME;
                        strftime(pom, sizeof(pom), "%a %H:%M %d.%m.%Y", localtime(&t));
                        UNLOCK_TIME;
                        strcpy(dp, pom);
                        while (*dp)
                        {
                            dp++;
                        }
                    }
                    break;

                default:
                    *dp = *p;
                    dp++;
                    p--;
                    break;
                }
            }
            else
            {
                *dp = *p;
                dp++;
            }
        }
        *dp = '\0';
    }
    else
    {
        snprintf(buf, sizeof(buf), "mailx %s -s \"pavuk reminder result\"\n", priv_cfg.from);
        buf[sizeof(buf) - 1] = 0;
    }

#ifdef HAVE_POPEN

    fp = popen(buf, "w");
    if (!fp)
    {
        xperror1("popen", buf);

        DEBUG_PROCE(NULL);
        return;
    }
    else
    {
        int cnt;
        dllist *ptr;

        fprintf(fp, gettext("This is the result of running pavuk reminder mode\n\n"));

        cnt = 0;
        for (ptr = remind_urls; ptr; ptr = ptr->next)
        {
            remind_entry *e = (remind_entry *)ptr->data;

            if (e->status & REMIND_MODIFIED)
            {
                if (!cnt)
                {
                    fprintf(fp, gettext("Changed URLs\n"));
                    fprintf(fp, "-------------------------\n");
                    cnt++;
                }
                p = url_to_urlstr(e->urlp, FALSE);
                LOCK_TIME;
                fprintf(fp, "%s %s", p, ctime(&e->mdtm));
                UNLOCK_TIME;
                _free(p);
            }
        }

        fprintf(fp, "\n\n");

        cnt = 0;
        for (ptr = remind_urls; ptr; ptr = ptr->next)
        {
            remind_entry *e = (remind_entry *)ptr->data;

            if (e->status & REMIND_ERROR)
            {
                if (!cnt)
                {
                    fprintf(fp, gettext("URLs with some errors\n"));
                    fprintf(fp, "-------------------------\n");
                    cnt++;
                }
                p = url_to_urlstr(e->urlp, FALSE);
                fprintf(fp, "%s\n", p);
                _free(p);
            }
        }

#ifdef HAVE_PCLOSE
        pclose(fp);
#endif
    }

#else /* HAVE_POPEN */

    fp = NULL;
    errno = ENOSUP;
    xperror1(gettext("popen not supported --> -remind_cmd not supported on this machine"), buf);

#endif

    DEBUG_PROCE(NULL);
}

