/***************************************************************************/
/*    This code is part of WWW grabber called pavuk                        */
/*    Copyright (c) 1997 - 2001 Stefan Ondrejicka                          */
/*    Distributed under GPL 2 or later                                     */
/***************************************************************************/

#ifndef _url_h_
#define _url_h_

#include "dllist.h"
#include "cookie.h"


typedef struct
{
    const char *host;                         /*** HTTP host address ***/
    unsigned short port;                      /*** HTTP service port number ***/
    const char *document;                     /*** document path ***/
    const char *searchstr;                    /*** query string ***/
    const char *anchor_name;                  /*** anchor name ***/
    const char *user;                         /*** username for authorization ***/
    const char *password;                     /*** password for authorization ***/
} url_http_t;

typedef struct
{
    const char *host;                         /*** FTP host address ***/
    unsigned short port;                      /*** FTP service port number ***/
    const char *user;                         /*** username for authorization ***/
    const char *password;                     /*** password for authorization ***/
    const char *path;                         /*** document path ***/
    const char *anchor_name;                  /*** anchor name ***/
    bool_t dir;                               /*** is this FTP directory URL ? ***/
} url_ftp_t;

typedef struct
{
    const char *filename;                     /*** file path ***/
    const char *searchstr;                    /*** query string ***/
    const char *anchor_name;                  /*** anchor name ***/
} url_file_t;

typedef struct
{
    const char *host;                         /*** GOPHER host address ***/
    unsigned short port;                      /*** GOPHER service port number ***/
    const char *selector;                     /*** document selector ***/
} url_gopher_t;

typedef struct
{
    int dummy;
} url_unsup_t;

/*
 * WARNING: prottable[] in url.c expects this particular order and implicit
 * numbering as several locations in the code use this enum as a direct index
 * into that table!
 */
typedef enum
{
    /*** id for URL types ***/
    URLT_UNKNOWN,
    URLT_HTTP,
    URLT_HTTPS,
    URLT_FTP,                   /* url->extension is optional (ftp_url_extension *) */
    URLT_FTPS,                  /* url->extension is optional (ftp_url_extension *) */
    URLT_FILE,
    URLT_GOPHER,
    URLT_FROMPARENT
} protocol;

/* refers to RFC 2396 */
#define URL_PATH_UNSAFE             " <>\"#%{}|\\^[]`+@=&$?;"
#define URL_QUERY_UNSAFE            " <>\"#%{}|\\^[]`"
#define URL_RQUERY_UNSAFE           " <>\"#%{}|\\^[]`+@=&$?;:/?,"
#define URL_AUTH_UNSAFE             " <>\"#%{}|\\^[]`+@=&$?;:/"


#define URL_REDIRECT            (unsigned int)(1 << 0)                          /* 0x00000001 */
#define URL_INLINE_OBJ          (unsigned int)(1 << 1)                          /* 0x00000002 */
#define URL_PROCESSED           (unsigned int)(1 << 2)                          /* 0x00000004 */
#define URL_DOWNLOADED          (unsigned int)(1 << 3)                          /* 0x00000008 */
#define URL_ERR_UNREC           (unsigned int)(1 << 4)                          /* 0x00000010 */
#define URL_MOVED               (unsigned int)(1 << 5)                          /* 0x00000020 */
#define URL_REJECTED            (unsigned int)(1 << 6)                          /* 0x00000040 */
#define URL_USER_DISABLED       (unsigned int)(1 << 7)                          /* 0x00000080 */
#define URL_NOT_FOUND           (unsigned int)(1 << 8)                          /* 0x00000100 */
#define URL_TRUNCATED           (unsigned int)(1 << 9)                          /* 0x00000200 */
#define URL_ERR_REC             (unsigned int)(1 << 10)                         /* 0x00000400 */
#define URL_STYLE               (unsigned int)(1 << 11)                         /* 0x00000800 */
#define URL_INNSCACHE           (unsigned int)(1 << 12)                         /* 0x00001000 */
#define URL_ISHTML              (unsigned int)(1 << 13)                         /* 0x00002000 */
#define URL_ISLOCAL             (unsigned int)(1 << 14)                         /* 0x00004000 */
#define URL_NORECURSE           (unsigned int)(1 << 15)                         /* 0x00008000 */
#define URL_FORM_ACTION         (unsigned int)(1 << 16) /* 0x00010000 */        /* url->extension is (form_info *) */
#define URL_HAVE_FORMS          (unsigned int)(1 << 17)                         /* 0x00020000 */
#define URL_ISFIRST             (unsigned int)(1 << 18)                         /* 0x00040000 */
#define URL_ISSTARTING          (unsigned int)(1 << 19)                         /* 0x00080000 */
#define URL_ISSCRIPT            (unsigned int)(1 << 20)                         /* 0x00100000 */
#ifdef INCLUDE_CHUNKY_DoS_FEATURES
#define URL_ISHAMMER            (unsigned int)(1 << 21)         /* 0x00200000 */
#endif
#define URL_PARENT_REWRITING    (unsigned int)(1 << 22)         /* 0x00400000 */


typedef struct _protinfo
{
    protocol id;
    const char *dirname;
    const char *urlid;
    const char *typestr;
    int default_port;
    bool_t supported;
} protinfo;

typedef struct
{
    /*** properties of document ***/
    const char *type;                         /*** MIME type of document ***/
    ssize_t size;                             /*** size of document ***/
    time_t mdtm;                              /*** modification time ***/
} url_prop;

typedef union
{
    url_http_t http;
    url_file_t file;
    url_ftp_t ftp;
    url_gopher_t gopher;
    url_unsup_t unsup;
} url_union_t;

typedef struct _url
{
    protocol type;                      /*** type of URL ***/

    dllist *parent_url;                 /*** list of parent URLs ***/
    struct _url *moved_to;              /*** pointer to new URL if document was moved ***/

    unsigned short level;               /*** tree level of document ***/
    unsigned short ref_cnt;             /*** number of references to this URL structure ***/
    unsigned int status;                /*** status flags of URL ***/

    const char *urlstr;                       /*** url string for unsupported type of URLs ***/
    url_union_t p;                            /*** parsed URL infos ***/
    const char *local_name;                   /*** assigned local filename ***/
    void *extension;                          /*** possible url extensions ***/

#ifdef WITH_TREE
#ifdef I_FACE
    url_prop *prop;                     /*** document properties ***/
    void **tree_nfo;                    /*** UI representation of tree nodes ***/
#endif
#endif                          /* WITH_TREE */
#ifdef HAVE_MT
    pthread_mutex_t lock;               /*** mt lock ***/
#endif
} url;

typedef enum
{
    URLI_UNKNOWN = 0,
    URLI_NORMAL = 1,
    URLI_FORM = 2
} uri_type;

typedef enum
{
    FORM_M_GET,
    FORM_M_POST,
    FORM_M_UNKNOWN
} form_method;

typedef enum
{
    FORM_E_MULTIPART,
    FORM_E_URLENCODED,
    FORM_E_UNKNOWN
} form_encoding;

typedef struct
{
    char *urlstr;
    uri_type type;
    form_method method;
    form_encoding encoding;
    dllist *fields;
    cookie_entry *cookies;      /* add these cookies when you hit this url */
    char *localname;
} url_info;

extern url_info *url_info_new(const char *urlstr);
extern url_info *url_info_parse(const char *str, bool_t try_me);
extern char *url_info_dump(url_info *);
extern void url_info_free(url_info *);
extern url_info *url_info_duplicate(url_info *);

extern url *url_parse(DBGdecl(const char *urlstr));
#define url_parse(u)    url_parse(DBGvars(u))
extern url *url_dup_url(DBGdecl(url *));
#define url_dup_url(u)  url_dup_url(DBGvars(u))
extern char *url_parse_scheme(const char *urlstr);
extern protocol url_scheme_to_schemeid(const char *scheme);
extern int dllist_url_compare(dllist_t key1, dllist_t key2);
extern int url_compare(url *u1, url *u2);
extern char *url_to_absolute_url(const char *base, const char *baset, url *parent, const char *act);
extern char *url_encode_str(const char *urlstr, const char *unsafe);
extern char *url_decode_str(const char *urlstr, int len);
extern url *new_url(DBGdecl(url *));
extern const char *url_to_filename(url *urlp, int lockfn);
extern const char *url_to_filename_with_type(url *urlp, const char *mime_type, int lockfn);
extern char *url_get_local_name_real(url *urlp, const char *mime_type, int adj);
extern void url_changed_filename(url *);
extern char *url_to_in_filename(url *);
extern void free_deep_url(url *);
extern char *url_to_urlstr(DBGdecl(url * urlp), int with_anchor);
#define url_to_urlstr(u, n)       url_to_urlstr(DBGvars(u), n)
extern char *url_to_request_urlstr(DBGdecl(url * urlp), int absolute);
#define url_to_request_urlstr(u, l)             url_to_request_urlstr(DBGvars(u), l)
extern void url_path_abs(DBGdecl(url *));
#define url_path_abs(u)         url_path_abs(DBGvars(u))
extern url *filename_to_url(const char *ifn);
extern void cat_links_to_url_list(dllist *);
extern void append_url_to_list(url *);
extern const char *url_get_site(url *urlr);
extern int url_get_port(url *);
extern const char *url_get_pass(url *urlr, const char *realm);
extern const char *url_get_user(url *urlr, const char *realm);
extern int url_get_auth_scheme(url *urlr, const char *realm);
extern const char *url_get_path(url *urlr);
extern char *url_get_full_path(url *urlr);
extern const char *url_get_anchor_name(url *urlp);
extern void url_clear_anchor(url *);
extern const char *url_get_search_str(url *urlp);
extern void url_set_path(url *urlr, const char *path);
extern int url_is_dir_index(url *);
extern int url_is_same_site(url *, url *);
extern void url_add_to_url_hash_tab(url *);
extern void url_remove_from_url_hash_tab(url *);
extern void url_add_to_file_hash_tab(url *);
extern void url_remove_from_file_hash_tab(url *);
extern url *url_was_befor(url *);
extern void link_url_in_list(url *orig, url *copy);
extern int url_redirect_to(url *src, url *dst, int is_303);
extern void url_forget_filename(url *);
extern void url_set_filename(url *urlp, const char *local_name);
extern char *cvt_url_info2diag_str(url_info *ui);
extern void cvt_url2diag_str(url *urlp, char **urldbgstr_ref, char **urlstr_ref);

extern const protinfo prottable[];
extern int prottable_size(void);

#endif

